# -*- coding: utf-8 -*-
"""SYSTEMTIME structure implementation."""

import decimal

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import interface


class Systemtime(interface.DateTimeValues):
  """SYSTEMTIME structure.

  The SYSTEMTIME structure is 16 bytes of size and contains:

  struct {
      WORD year,
      WORD month,
      WORD day_of_week,
      WORD day_of_month,
      WORD hour,
      WORD minute,
      WORD second,
      WORD millisecond
  }
  """

  def __init__(
      self, precision=None, system_time_tuple=None, time_zone_offset=None):
    """Initializes a SYSTEMTIME structure.

    Args:
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      system_time_tuple
          (Optional[tuple[int, int, int, int, int, int, int, int]]):
          system time, contains year, month, day of week, day of month,
          hours, minutes, seconds and milliseconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the system time is invalid.
    """
    super(Systemtime, self).__init__(
        precision=precision or definitions.PRECISION_1_MILLISECOND,
        time_zone_offset=time_zone_offset)
    self._number_of_seconds = None
    self._day_of_month = None
    self._day_of_week = None
    self._hours = None
    self._milliseconds = None
    self._minutes = None
    self._month = None
    self._seconds = None
    self._year = None

    if system_time_tuple:
      if len(system_time_tuple) < 8:
        raise ValueError('Invalid system time tuple 8 elements required.')

      if system_time_tuple[0] < 1601 or system_time_tuple[0] > 30827:
        raise ValueError('Year value out of bounds.')

      if system_time_tuple[1] not in range(1, 13):
        raise ValueError('Month value out of bounds.')

      if system_time_tuple[2] not in range(0, 7):
        raise ValueError('Day of week value out of bounds.')

      days_per_month = self._GetDaysPerMonth(
          system_time_tuple[0], system_time_tuple[1])
      if system_time_tuple[3] < 1 or system_time_tuple[3] > days_per_month:
        raise ValueError('Day of month value out of bounds.')

      if system_time_tuple[4] not in range(0, 24):
        raise ValueError('Hours value out of bounds.')

      if system_time_tuple[5] not in range(0, 60):
        raise ValueError('Minutes value out of bounds.')

      # TODO: support a leap second?
      if system_time_tuple[6] not in range(0, 60):
        raise ValueError('Seconds value out of bounds.')

      if system_time_tuple[7] < 0 or system_time_tuple[7] > 999:
        raise ValueError('Milliseconds value out of bounds.')

      self._day_of_month = system_time_tuple[3]
      self._day_of_week = system_time_tuple[2]
      self._hours = system_time_tuple[4]
      self._milliseconds = system_time_tuple[7]
      self._minutes = system_time_tuple[5]
      self._month = system_time_tuple[1]
      self._seconds = system_time_tuple[6]
      self._year = system_time_tuple[0]

      self._number_of_seconds = self._GetNumberOfSecondsFromElements(
          self._year, self._month, self._day_of_month, self._hours,
          self._minutes, self._seconds)

  @property
  def day_of_month(self):
    """day_of_month (int): day of month, 1 through 31."""
    return self._day_of_month

  @property
  def day_of_week(self):
    """day_of_week (int): day of week, 0 through 6."""
    return self._day_of_week

  @property
  def hours(self):
    """Hours (int): hours, 0 through 23."""
    return self._hours

  @property
  def milliseconds(self):
    """Milliseconds (int): milliseconds, 0 through 999."""
    return self._milliseconds

  @property
  def minutes(self):
    """Minutes (int): minutes, 0 through 59."""
    return self._minutes

  @property
  def month(self):
    """Month (int): month of year, 1 through 12."""
    return self._month

  @property
  def seconds(self):
    """Seconds (int): seconds, 0 through 59."""
    return self._seconds

  @property
  def year(self):
    """Year (int): year, 1601 through 30827."""
    return self._year

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._number_of_seconds is not None:
        self._normalized_timestamp = (
            decimal.Decimal(self._milliseconds) /
            definitions.MILLISECONDS_PER_SECOND)
        self._normalized_timestamp += decimal.Decimal(self._number_of_seconds)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def CopyFromDateTimeString(self, time_string):
    """Copies a SYSTEMTIME structure from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the date string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    milliseconds, _ = divmod(
        nanoseconds, definitions.NANOSECONDS_PER_MILLISECOND)

    if year < 1601 or year > 30827:
      raise ValueError(f'Unsupported year value: {year:d}.')

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._time_zone_offset = time_zone_offset

    self._year = year
    self._month = month
    self._day_of_month = day_of_month
    # TODO: calculate day of week on demand.
    self._day_of_week = None
    self._hours = hours
    self._minutes = minutes
    self._seconds = seconds
    self._milliseconds = milliseconds

  def CopyToDateTimeString(self):
    """Copies the SYSTEMTIME structure to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.###" or
          None if the number of seconds is missing.
    """
    if self._number_of_seconds is None:
      return None

    return (f'{self._year:04d}-{self._month:02d}-{self._day_of_month:02d} '
            f'{self._hours:02d}:{self._minutes:02d}:{self._seconds:02d}'
            f'.{self._milliseconds:03d}')


factory.Factory.RegisterDateTimeValues(Systemtime)



# -*- coding: utf-8 -*-
"""Digital Forensics Date and Time (dfDateTime).

dfDateTime, or Digital Forensics date and time, provides date and time objects
to preserve accuracy and precision.
"""

# Imports for date time values factory.
from dfdatetime import apfs_time
from dfdatetime import cocoa_time
from dfdatetime import delphi_date_time
from dfdatetime import dotnet_datetime
from dfdatetime import fat_date_time
from dfdatetime import filetime
from dfdatetime import hfs_time
from dfdatetime import golang_time
from dfdatetime import java_time
from dfdatetime import ole_automation_date
from dfdatetime import posix_time
from dfdatetime import rfc2579_date_time
from dfdatetime import semantic_time
from dfdatetime import systemtime
from dfdatetime import time_elements
from dfdatetime import uuid_time
from dfdatetime import webkit_time


__version__ = '20260411'



# -*- coding: utf-8 -*-
"""The date and time values serializer."""

from dfdatetime import factory
from dfdatetime import interface


class Serializer(object):
  """Date and time values serializer."""

  @classmethod
  def ConvertDictToDateTimeValues(cls, json_dict):
    """Converts a JSON dict into a date time values object.

    This method is deprecated use ConvertJSONToDateTimeValues instead.

    The dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'DateTimeValues'
        '__class_name__': 'RFC2579DateTime'
        ...
    }

    Here '__type__' indicates the object base type. In this case this should
    be 'DateTimeValues'. The rest of the elements of the dictionary make up the
    date time values object properties.

    Args:
      json_dict (dict[str, object]): JSON serialized objects.

    Returns:
      dfdatetime.DateTimeValues: date and time values.
    """
    return cls.ConvertJSONToDateTimeValues(json_dict)

  @classmethod
  def ConvertDateTimeValuesToDict(cls, date_time_values):
    """Converts a date and time values object into a JSON dictionary.

    This method is deprecated use ConvertDateTimeValuesToJSON instead.

    The resulting dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'DateTimeValues'
        '__class_name__': 'RFC2579DateTime'
        ...
    }

    Here '__type__' indicates the object base type. In this case
    'DateTimeValues'. The rest of the elements of the dictionary make up the
    date and time value object properties.

    Args:
      date_time_values (dfdatetime.DateTimeValues): date and time values.

    Returns:
      dict[str, object]: JSON serialized objects.

    Raises:
      TypeError: if object is not an instance of DateTimeValues.
    """
    if not isinstance(date_time_values, interface.DateTimeValues):
      raise TypeError

    return cls.ConvertDateTimeValuesToJSON(date_time_values)

  @classmethod
  def ConvertDateTimeValuesToJSON(cls, date_time_values):
    """Converts a date and time values object into a JSON dictionary.

    The resulting dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'DateTimeValues'
        '__class_name__': 'RFC2579DateTime'
        ...
    }

    Here '__type__' indicates the object base type. In this case
    'DateTimeValues'. The rest of the elements of the dictionary make up the
    date and time value object properties.

    Args:
      date_time_values (dfdatetime.DateTimeValues): date and time values.

    Returns:
      dict[str, object]: JSON serialized objects.
    """
    class_name = type(date_time_values).__name__

    json_dict = {
        '__class_name__': class_name,
        '__type__': 'DateTimeValues'}

    if hasattr(date_time_values, 'timestamp'):
      json_dict['timestamp'] = date_time_values.timestamp

    elif hasattr(date_time_values, 'string'):
      json_dict['string'] = date_time_values.string

    elif class_name == 'FATDateTime':
      json_dict['fat_date_time'] = date_time_values.fat_date_time

    elif class_name == 'GolangTime':
      json_dict['golang_timestamp'] = date_time_values.golang_timestamp

    elif class_name == 'RFC2579DateTime':
      time_zone_hours, time_zone_minutes = divmod(
          date_time_values.time_zone_offset, 60)

      if date_time_values.time_zone_offset < 0:
        time_zone_sign = '-'
        time_zone_hours *= -1
      else:
        time_zone_sign = '+'

      json_dict['rfc2579_date_time_tuple'] = (
          date_time_values.year, date_time_values.month,
          date_time_values.day_of_month, date_time_values.hours,
          date_time_values.minutes, date_time_values.seconds,
          date_time_values.deciseconds, time_zone_sign, time_zone_hours,
          time_zone_minutes)

    elif class_name == 'Systemtime':
      json_dict['system_time_tuple'] = (
          date_time_values.year, date_time_values.month,
          date_time_values.day_of_week, date_time_values.day_of_month,
          date_time_values.hours, date_time_values.minutes,
          date_time_values.seconds, date_time_values.milliseconds)

    elif class_name == 'TimeElements':
      json_dict['time_elements_tuple'] = (
          date_time_values.year, date_time_values.month,
          date_time_values.day_of_month, date_time_values.hours,
          date_time_values.minutes, date_time_values.seconds)

    elif class_name == 'TimeElementsInMilliseconds':
      json_dict['time_elements_tuple'] = (
          date_time_values.year, date_time_values.month,
          date_time_values.day_of_month, date_time_values.hours,
          date_time_values.minutes, date_time_values.seconds,
          date_time_values.milliseconds)

    elif class_name == 'TimeElementsInMicroseconds':
      json_dict['time_elements_tuple'] = (
          date_time_values.year, date_time_values.month,
          date_time_values.day_of_month, date_time_values.hours,
          date_time_values.minutes, date_time_values.seconds,
          date_time_values.microseconds)

    if date_time_values.time_zone_offset is not None and class_name not in (
        'GolangTime', 'RFC2579DateTime'):
      json_dict['time_zone_offset'] = date_time_values.time_zone_offset

    if date_time_values.is_delta and class_name in (
        'TimeElements', 'TimeElementsInMilliseconds',
        'TimeElementsInMicroseconds'):
      json_dict['is_delta'] = True

    if date_time_values.is_local_time:
      json_dict['is_local_time'] = True
    if date_time_values.time_zone_hint:
      json_dict['time_zone_hint'] = date_time_values.time_zone_hint

    return json_dict

  @classmethod
  def ConvertJSONToDateTimeValues(cls, json_dict):
    """Converts a JSON dict into a date time values object.

    The dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'DateTimeValues'
        '__class_name__': 'RFC2579DateTime'
        ...
    }

    Here '__type__' indicates the object base type. In this case this should
    be 'DateTimeValues'. The rest of the elements of the dictionary make up the
    date time values object properties.

    Args:
      json_dict (dict[str, object]): JSON serialized objects.

    Returns:
      dfdatetime.DateTimeValues: date and time values.
    """
    class_name = json_dict.get('__class_name__', None)
    if class_name:
      del json_dict['__class_name__']

    # Remove the class type from the JSON dict since we cannot pass it.
    del json_dict['__type__']

    if class_name not in (
        'TimeElements', 'TimeElementsInMilliseconds',
        'TimeElementsInMicroseconds'):
      is_delta = json_dict.get('is_delta', None)
      if is_delta is not None:
        del json_dict['is_delta']

    is_local_time = json_dict.get('is_local_time', None)
    if is_local_time is not None:
      del json_dict['is_local_time']

    time_zone_hint = json_dict.get('time_zone_hint', None)
    if time_zone_hint is not None:
      del json_dict['time_zone_hint']

    if class_name in ('InvalidTime', 'Never', 'NotSet'):
      string = json_dict.get('string', None)
      if string is not None:
        del json_dict['string']

    if class_name in ('GolangTime', 'RFC2579DateTime'):
      time_zone_offset = json_dict.get('time_zone_offset', None)
      if time_zone_offset is not None:
        del json_dict['time_zone_offset']

    date_time = factory.Factory.NewDateTimeValues(class_name, **json_dict)
    if is_local_time:
      date_time.is_local_time = is_local_time
    if time_zone_hint:
      date_time.time_zone_hint = time_zone_hint

    return date_time



# -*- coding: utf-8 -*-
"""RFC2579 date-time implementation."""

import decimal

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import interface


class RFC2579DateTime(interface.DateTimeValues):
  """RFC2579 date-time.

  The RFC2579 date-time structure is 11 bytes of size and contains:

  struct {
      uin16_t year,
      uint8_t month,
      uint8_t day_of_month,
      uint8_t hours,
      uint8_t minutes,
      uint8_t seconds,
      uint8_t deciseconds,
      char direction_from_utc,
      uint8_t hours_from_utc,
      uint8_t minutes_from_utc
  }

  Also see:
    https://datatracker.ietf.org/doc/html/rfc2579

  Attributes:
    year (int): year, 0 through 65536.
    month (int): month of year, 1 through 12.
    day_of_month (int): day of month, 1 through 31.
    hours (int): hours, 0 through 23.
    minutes (int): minutes, 0 through 59.
    seconds (int): seconds, 0 through 59, where 60 is used to represent
        a leap-second.
    deciseconds (int): deciseconds, 0 through 9.
  """

  # TODO: make attributes read-only.

  # pylint: disable=missing-type-doc

  def __init__(self, precision=None, rfc2579_date_time_tuple=None):
    """Initializes a RFC2579 date-time.

    Args:
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      rfc2579_date_time_tuple:
          (Optional[tuple[int, int, int, int, int, int, int, str, int, int]]):
          RFC2579 date-time time, contains year, month, day of month, hours,
          minutes, seconds and deciseconds, and time zone offset in hours and
          minutes from UTC.

    Raises:
      ValueError: if the system time is invalid.
    """
    super(RFC2579DateTime, self).__init__(
        precision=precision or definitions.PRECISION_100_MILLISECONDS)
    self._day_of_month = None
    self._deciseconds = None
    self._hours = None
    self._minutes = None
    self._month = None
    self._number_of_seconds = None
    self._seconds = None
    self._year = None

    if rfc2579_date_time_tuple:
      if len(rfc2579_date_time_tuple) < 10:
        raise ValueError(
            'Invalid RFC2579 date-time tuple 10 elements required.')

      if rfc2579_date_time_tuple[0] < 0 or rfc2579_date_time_tuple[0] > 65536:
        raise ValueError('Year value out of bounds.')

      if rfc2579_date_time_tuple[1] not in range(1, 13):
        raise ValueError('Month value out of bounds.')

      days_per_month = self._GetDaysPerMonth(
          rfc2579_date_time_tuple[0], rfc2579_date_time_tuple[1])
      if (rfc2579_date_time_tuple[2] < 1 or
          rfc2579_date_time_tuple[2] > days_per_month):
        raise ValueError('Day of month value out of bounds.')

      if rfc2579_date_time_tuple[3] not in range(0, 24):
        raise ValueError('Hours value out of bounds.')

      if rfc2579_date_time_tuple[4] not in range(0, 60):
        raise ValueError('Minutes value out of bounds.')

      # TODO: support a leap second?
      if rfc2579_date_time_tuple[5] not in range(0, 60):
        raise ValueError('Seconds value out of bounds.')

      if rfc2579_date_time_tuple[6] < 0 or rfc2579_date_time_tuple[6] > 9:
        raise ValueError('Deciseconds value out of bounds.')

      if rfc2579_date_time_tuple[7] not in ('+', '-'):
        raise ValueError('Direction from UTC value out of bounds.')

      if rfc2579_date_time_tuple[8] not in range(0, 14):
        raise ValueError('Hours from UTC value out of bounds.')

      if rfc2579_date_time_tuple[9] not in range(0, 60):
        raise ValueError('Minutes from UTC value out of bounds.')

      time_zone_offset = (
          (rfc2579_date_time_tuple[8] * 60) + rfc2579_date_time_tuple[9])

      if rfc2579_date_time_tuple[7] == '-':
        time_zone_offset = -time_zone_offset

      self._time_zone_offset = time_zone_offset

      self._year = rfc2579_date_time_tuple[0]
      self._month = rfc2579_date_time_tuple[1]
      self._day_of_month = rfc2579_date_time_tuple[2]
      self._hours = rfc2579_date_time_tuple[3]
      self._minutes = rfc2579_date_time_tuple[4]
      self._seconds = rfc2579_date_time_tuple[5]
      self._deciseconds = rfc2579_date_time_tuple[6]

      self._number_of_seconds = self._GetNumberOfSecondsFromElements(
          self._year, self._month, self._day_of_month, self._hours,
          self._minutes, self._seconds)

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._number_of_seconds is not None:
        self._normalized_timestamp = (
            decimal.Decimal(self._deciseconds) /
            definitions.DECISECONDS_PER_SECOND)
        self._normalized_timestamp += decimal.Decimal(self._number_of_seconds)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  @property
  def deciseconds(self):
    """int: number of deciseconds or None if not set."""
    return self._deciseconds

  @property
  def day_of_month(self):
    """int: day of month or None if not set."""
    return self._day_of_month

  @property
  def hours(self):
    """int: number of hours or None if not set."""
    return self._hours

  @property
  def minutes(self):
    """int: number of minutes or None if not set."""
    return self._minutes

  @property
  def month(self):
    """int: month or None if not set."""
    return self._month

  @property
  def seconds(self):
    """int: number of seconds or None if not set."""
    return self._seconds

  @property
  def year(self):
    """int: year or None if not set."""
    return self._year

  def CopyFromDateTimeString(self, time_string):
    """Copies a RFC2579 date-time from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the date string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    deciseconds, _ = divmod(
        nanoseconds, definitions.NANOSECONDS_PER_DECISECOND)

    if year < 0 or year > 65536:
      raise ValueError(f'Unsupported year value: {year:d}.')

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._time_zone_offset = time_zone_offset

    self._year = year
    self._month = month
    self._day_of_month = day_of_month
    self._hours = hours
    self._minutes = minutes
    self._seconds = seconds
    self._deciseconds = deciseconds

  def CopyToDateTimeString(self):
    """Copies the RFC2579 date-time to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.#" or
          None if the number of seconds is missing.
    """
    if self._number_of_seconds is None:
      return None

    return (f'{self._year:04d}-{self._month:02d}-{self._day_of_month:02d} '
            f'{self._hours:02d}:{self._minutes:02d}:{self._seconds:02d}'
            f'.{self._deciseconds:01d}')


factory.Factory.RegisterDateTimeValues(RFC2579DateTime)



# -*- coding: utf-8 -*-
"""The date and time values factory."""


class Factory(object):
  """Date and time values factory."""

  _date_time_values_types = {}

  @classmethod
  def DeregisterDateTimeValues(cls, date_time_values_type):
    """Deregisters a date and time values type.

    Args:
      date_time_values_type (type): date and time values type.

    Raises:
      KeyError: if date and time values type is not registered.
    """
    class_name = date_time_values_type.__name__
    if class_name not in cls._date_time_values_types:
      raise KeyError(f'Date and time values type: {class_name:s} not set.')

    del cls._date_time_values_types[class_name]

  @classmethod
  def NewDateTimeValues(cls, class_name, **kwargs):
    """Creates a new date and time values for the specific type indicator.

    Args:
      class_name (str): type indicator.
      kwargs (dict): keyword arguments depending on the date and time values.

    Returns:
      DateTimeValues: date and time values.

    Raises:
      KeyError: if date and time values is not registered.
    """
    if class_name not in cls._date_time_values_types:
      raise KeyError(f'Date and time values type: {class_name:s} not set.')

    date_time_values_type = cls._date_time_values_types[class_name]
    return date_time_values_type(**kwargs)

  @classmethod
  def RegisterDateTimeValues(cls, date_time_values_type):
    """Registers a date and time values type.

    Args:
      date_time_values_type (type): date and time values type.

    Raises:
      KeyError: if date and time values is already registered.
    """
    class_name = date_time_values_type.__name__
    if class_name in cls._date_time_values_types:
      raise KeyError(f'Date and time values type: {class_name:s} already set.')

    cls._date_time_values_types[class_name] = date_time_values_type



# -*- coding: utf-8 -*-
"""Time elements implementation."""

import decimal

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import interface
from dfdatetime import precisions


class TimeElements(interface.DateTimeValues):
  """Time elements.

  Time elements contain separate values for year, month, day of month,
  hours, minutes and seconds.

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
  """

  # Maps the RFC 822, RFC 1123 and RFC 2822 definitions to their corresponding
  # integer values.
  _RFC_MONTH_MAPPINGS = {
      'Jan': 1,
      'Feb': 2,
      'Mar': 3,
      'Apr': 4,
      'May': 5,
      'Jun': 6,
      'Jul': 7,
      'Aug': 8,
      'Sep': 9,
      'Oct': 10,
      'Nov': 11,
      'Dec': 12}

  _RFC_TIME_ZONE_MAPPINGS = {
      'UT': 0,
      'GMT':  0,
      'EST': -5,
      'EDT': -4,
      'CST': -6,
      'CDT': -5,
      'MST': -7,
      'MDT': -6,
      'PST': -8,
      'PDT': -7,
      'A': -1,
      'B': -2,
      'C': -3,
      'D': -4,
      'E': -5,
      'F': -6,
      'G': -7,
      'H': -8,
      'I': -9,
      'K': -10,
      'L': -11,
      'M': -12,
      'N': 1,
      'O': 2,
      'P': 3,
      'Q': 4,
      'R': 5,
      'S': 6,
      'T': 7,
      'U': 8,
      'V': 9,
      'W': 10,
      'X': 11,
      'Y': 12,
      'Z': 0}

  _RFC_WEEKDAYS = frozenset(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'])

  def __init__(
      self, is_delta=False, precision=None, time_elements_tuple=None,
      time_zone_offset=None):
    """Initializes time elements.

    Args:
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_elements_tuple (Optional[tuple[int, int, int, int, int, int]]):
          time elements, contains year, month, day of month, hours, minutes and
          seconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    super(TimeElements, self).__init__(
        is_delta=is_delta,
        precision=precision or definitions.PRECISION_1_SECOND,
        time_zone_offset=time_zone_offset)
    self._number_of_seconds = None
    self._time_elements_tuple = time_elements_tuple

    if time_elements_tuple:
      number_of_elements = len(time_elements_tuple)
      if number_of_elements < 6:
        raise ValueError((
            f'Invalid time elements tuple at least 6 elements required,'
            f'got: {number_of_elements:d}'))

      self._number_of_seconds = self._GetNumberOfSecondsFromElements(
          time_elements_tuple[0], time_elements_tuple[1],
          time_elements_tuple[2], time_elements_tuple[3],
          time_elements_tuple[4], time_elements_tuple[5])

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._number_of_seconds is not None:
        self._normalized_timestamp = decimal.Decimal(self._number_of_seconds)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def _CopyDateTimeFromStringISO8601(self, time_string):
    """Copies a date and time from an ISO 8601 date and time string.

    Args:
      time_string (str): time value formatted as:
          hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The fraction of second and
          time zone offset are optional.

    Returns:
      dict[str, int]: date and time values, such as year, month, day of month,
          hours, minutes, seconds, nanoseconds, time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    if not time_string:
      raise ValueError('Invalid time string.')

    time_string_length = len(time_string)

    year, month, day_of_month = self._CopyDateFromString(time_string)

    if time_string_length <= 10:
      return {
          'year': year,
          'month': month,
          'day_of_month': day_of_month}

    # If a time of day is specified the time string it should at least
    # contain 'YYYY-MM-DDThh'.
    if time_string[10] != 'T':
      raise ValueError('Invalid time string - missing date and time separator.')

    hours, minutes, seconds, nanoseconds, time_zone_offset = (
        self._CopyTimeFromStringISO8601(time_string[11:]))

    date_time_values = {
        'year': year,
        'month': month,
        'day_of_month': day_of_month,
        'hours': hours,
        'minutes': minutes,
        'seconds': seconds}

    if nanoseconds is not None:
      date_time_values['nanoseconds'] = nanoseconds
    if time_zone_offset is not None:
      date_time_values['time_zone_offset'] = time_zone_offset

    return date_time_values

  def _CopyDateTimeFromStringRFC822(self, time_string):
    """Copies a date and time from a RFC 822 date and time string.

    Args:
      time_string (str): date and time value formatted as:
          DAY, D MONTH YY hh:mm:ss ZONE

          Where weekday (DAY) and seconds (ss) are optional and day of
          month (D) can consist of 1 or 2 digits.

    Returns:
      dict[str, int]: date and time values, such as year, month, day of month,
          hours, minutes, seconds, time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    if not time_string:
      raise ValueError('Invalid time string.')

    string_segments = time_string.split(' ')

    if len(string_segments) not in (5, 6):
      raise ValueError('Unsupported number of time string segments.')

    weekday_string = string_segments[0]
    if weekday_string.endswith(','):
      weekday_string = weekday_string[:-1]
      if weekday_string not in self._RFC_WEEKDAYS:
        raise ValueError(f'Invalid weekday: {weekday_string:s}.')

      string_segments.pop(0)

    day_of_month_string = string_segments[0]

    day_of_month = 0
    if len(day_of_month_string) in (1, 2):
      try:
        day_of_month = int(day_of_month_string, 10)
      except ValueError:
        pass

    if day_of_month == 0:
      raise ValueError(f'Invalid day of month: {day_of_month_string:s}.')

    month_string = string_segments[1]

    month = self._RFC_MONTH_MAPPINGS.get(month_string)
    if not month:
      raise ValueError(f'Invalid month: {month_string:s}.')

    year_string = string_segments[2]

    year = None
    if len(year_string) == 2:
      try:
        year = int(year_string, 10)
      except ValueError:
        pass

    if year is None:
      raise ValueError(f'Invalid year: {0:s}.')

    year += 1900

    hours, minutes, seconds, time_zone_offset = self._CopyTimeFromStringRFC(
        string_segments[3], string_segments[4])

    date_time_values = {
        'year': year,
        'month': month,
        'day_of_month': day_of_month,
        'hours': hours,
        'minutes': minutes,
        'time_zone_offset': time_zone_offset}

    if seconds is not None:
      date_time_values['seconds'] = seconds

    return date_time_values

  def _CopyDateTimeFromStringRFC1123(self, time_string):
    """Copies a date and time from a RFC 1123 date and time string.

    Args:
      time_string (str): date and time value formatted as:
          DAY, D MONTH YYYY hh:mm:ss ZONE

          Where weekday (DAY) and seconds (ss) are optional and day of
          month (D) can consist of 1 or 2 digits.

    Returns:
      dict[str, int]: date and time values, such as year, month, day of month,
          hours, minutes, seconds, time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    if not time_string:
      raise ValueError('Invalid time string.')

    string_segments = time_string.split(' ')

    if len(string_segments) not in (5, 6):
      raise ValueError('Unsupported number of time string segments.')

    weekday_string = string_segments[0]
    if weekday_string.endswith(','):
      weekday_string = weekday_string[:-1]
      if weekday_string not in self._RFC_WEEKDAYS:
        raise ValueError(f'Invalid weekday: {weekday_string:s}.')

      string_segments.pop(0)

    day_of_month_string = string_segments[0]

    day_of_month = 0
    if len(day_of_month_string) in (1, 2):
      try:
        day_of_month = int(day_of_month_string, 10)
      except ValueError:
        pass

    if day_of_month == 0:
      raise ValueError(f'Invalid day of month: {day_of_month_string:s}.')

    month_string = string_segments[1]

    month = self._RFC_MONTH_MAPPINGS.get(month_string)
    if not month:
      raise ValueError(f'Invalid month: {month_string:s}.')

    year_string = string_segments[2]

    year = None
    if len(year_string) == 4:
      try:
        year = int(year_string, 10)
      except ValueError:
        pass

    if year is None:
      raise ValueError(f'Invalid year: {year_string:s}.')

    hours, minutes, seconds, time_zone_offset = self._CopyTimeFromStringRFC(
        string_segments[3], string_segments[4])

    date_time_values = {
        'year': year,
        'month': month,
        'day_of_month': day_of_month,
        'hours': hours,
        'minutes': minutes,
        'time_zone_offset': time_zone_offset}

    if seconds is not None:
      date_time_values['seconds'] = seconds

    return date_time_values

  def _CopyFromDateTimeValues(self, date_time_values):
    """Copies time elements from date and time values.

    Args:
      date_time_values  (dict[str, int]): date and time values, such as year,
          month, day of month, hours, minutes, seconds, nanoseconds, time zone
          offset in minutes.
    """
    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds)
    self._time_zone_offset = time_zone_offset

  def _CopyTimeFromStringISO8601(self, time_string):
    """Copies a time from an ISO 8601 time string.

    Args:
      time_string (str): time value formatted as:
          hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The fraction of second and
          time zone offset are optional.

    Returns:
      tuple[int, int, int, int, int]: hours, minutes, seconds, nanoseconds,
          time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    if time_string.endswith('Z'):
      time_string = ''.join([time_string[:-1], '+00:00'])

    time_string_length = len(time_string)

    # The time string should at least contain 'hh'.
    if time_string_length < 2:
      raise ValueError('Time string too short.')

    try:
      hours = int(time_string[0:2], 10)
    except ValueError:
      raise ValueError('Unable to parse hours.')

    if hours not in range(0, 24):
      raise ValueError(f'Hours value: {hours:d} out of bounds.')

    minutes = None
    seconds = None
    nanoseconds = None
    time_zone_offset = None

    time_string_index = 2

    # Minutes are either specified as 'hhmm', 'hh:mm' or as a fractional part
    # 'hh[.,]###'.
    if (time_string_index + 1 < time_string_length and
        time_string[time_string_index] not in ('.', ',')):
      if time_string[time_string_index] == ':':
        time_string_index += 1

      if time_string_index + 2 > time_string_length:
        raise ValueError('Time string too short.')

      try:
        minutes = time_string[time_string_index:time_string_index + 2]
        minutes = int(minutes, 10)
      except ValueError:
        raise ValueError('Unable to parse minutes.')

      time_string_index += 2

    # Seconds are either specified as 'hhmmss', 'hh:mm:ss' or as a fractional
    # part 'hh:mm[.,]###' or 'hhmm[.,]###'.
    if (time_string_index + 1 < time_string_length and
        time_string[time_string_index] not in ('.', ',')):
      if time_string[time_string_index] == ':':
        time_string_index += 1

      if time_string_index + 2 > time_string_length:
        raise ValueError('Time string too short.')

      try:
        seconds = time_string[time_string_index:time_string_index + 2]
        seconds = int(seconds, 10)
      except ValueError:
        raise ValueError('Unable to parse day of seconds.')

      time_string_index += 2

    time_zone_string_index = time_string_index
    while time_zone_string_index < time_string_length:
      if time_string[time_zone_string_index] in ('+', '-'):
        break

      time_zone_string_index += 1

    # The calculations that follow rely on the time zone string index
    # to point beyond the string in case no time zone offset was defined.
    if time_zone_string_index == time_string_length - 1:
      time_zone_string_index += 1

    if (time_string_length > time_string_index and
        time_string[time_string_index] in ('.', ',')):
      time_string_index += 1
      time_fraction_length = time_zone_string_index - time_string_index

      try:
        time_fraction = time_string[time_string_index:time_zone_string_index]
        time_fraction = int(time_fraction, 10)
        time_fraction = (
            decimal.Decimal(time_fraction) /
            decimal.Decimal(10 ** time_fraction_length))
      except ValueError:
        raise ValueError('Unable to parse time fraction.')

      if minutes is None:
        time_fraction *= 60
        minutes = int(time_fraction)
        time_fraction -= minutes

      if seconds is None:
        time_fraction *= 60
        seconds = int(time_fraction)
        time_fraction -= seconds

      time_fraction *= definitions.NANOSECONDS_PER_SECOND
      nanoseconds = int(time_fraction)

    if minutes is not None and minutes not in range(0, 60):
      raise ValueError(f'Minutes value: {minutes:d} out of bounds.')

    # TODO: support a leap second?
    if seconds is not None and seconds not in range(0, 60):
      raise ValueError(f'Seconds value: {seconds:d} out of bounds.')

    if time_zone_string_index < time_string_length:
      if (time_string_length - time_zone_string_index != 6 or
          time_string[time_zone_string_index + 3] != ':'):
        raise ValueError('Invalid time string.')

      try:
        hours_from_utc = int(time_string[
            time_zone_string_index + 1:time_zone_string_index + 3])
      except ValueError:
        raise ValueError('Unable to parse time zone hours offset.')

      if hours_from_utc not in range(0, 15):
        raise ValueError('Time zone hours offset value out of bounds.')

      try:
        minutes_from_utc = int(time_string[
            time_zone_string_index + 4:time_zone_string_index + 6])
      except ValueError:
        raise ValueError('Unable to parse time zone minutes offset.')

      if minutes_from_utc not in range(0, 60):
        raise ValueError('Time zone minutes offset value out of bounds.')

      # pylint: disable=invalid-unary-operand-type
      time_zone_offset = (hours_from_utc * 60) + minutes_from_utc

      if time_string[time_zone_string_index] == '-':
        time_zone_offset = -time_zone_offset

    return hours, minutes, seconds, nanoseconds, time_zone_offset

  def _CopyTimeFromStringRFC(self, time_string, time_zone_string):
    """Copies a time from a RFC 822, RFC 1123 or RFC 2822 time string.

    Args:
      time_string (str): time value formatted as: hh:mm[:ss], where seconds (ss)
          are optional.
      time_zone_string (str): time zone value formatted as predefined time zone
          indicator or [+-]HHMM

    Returns:
      tuple[int, int, int, int]: hours, minutes, seconds, time zone offset in
          minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    time_string_length = len(time_string)

    # The time string should at least contain 'hh:mm'.
    if time_string_length < 5:
      raise ValueError('Time string too short.')

    if time_string_length > 8:
      raise ValueError('Time string too long.')

    if time_string[2] != ':':
      raise ValueError('Invalid hours and minutes separator.')

    try:
      hours = int(time_string[0:2], 10)
    except ValueError:
      raise ValueError('Unable to parse hours.')

    if hours not in range(0, 24):
      raise ValueError(f'Hours value: {hours:d} out of bounds.')

    try:
      minutes = int(time_string[3:5], 10)
    except ValueError:
      raise ValueError('Unable to parse minutes.')

    if minutes not in range(0, 60):
      raise ValueError(f'Minutes value: {minutes:d} out of bounds.')

    seconds = None

    if time_string_length > 5:
      if time_string_length < 8:
        raise ValueError('Time string too short.')

      if time_string[5] != ':':
        raise ValueError('Invalid minutes and seconds separator.')

      try:
        seconds = int(time_string[6:8], 10)
      except ValueError:
        raise ValueError('Unable to parse seconds.')

      if seconds not in range(0, 60):
        raise ValueError(f'Seconds value: {seconds:d} out of bounds.')

    if time_string_length < 5:
      raise ValueError('Time string too short.')

    time_zone_string_length = len(time_zone_string)
    if time_zone_string_length > 5:
      raise ValueError('Time zone string too long.')

    if time_zone_string_length < 5:
      hours_from_utc = self._RFC_TIME_ZONE_MAPPINGS.get(time_zone_string, None)
      minutes_from_utc = 0
      if hours_from_utc is None:
        raise ValueError(f'Invalid time zone: {time_zone_string:s}.')

    else:
      if time_zone_string[0] not in ('+', '-'):
        raise ValueError(f'Invalid time zone: {time_zone_string:s}.')

      try:
        hours_from_utc = int(time_zone_string[1:3], 10)
      except ValueError:
        raise ValueError('Unable to parse time zone hours offset.')

      if hours_from_utc not in range(0, 15):
        raise ValueError('Time zone hours offset value out of bounds.')

      try:
        minutes_from_utc = int(time_zone_string[3:5], 10)
      except ValueError:
        raise ValueError('Unable to parse time zone minutes offset.')

      if minutes_from_utc not in range(0, 60):
        raise ValueError('Time zone minutes offset value out of bounds.')

    time_zone_offset = (hours_from_utc * 60) + minutes_from_utc
    if time_zone_string[0] == '-':
      time_zone_offset = -time_zone_offset

    return hours, minutes, seconds, time_zone_offset

  @property
  def day_of_month(self):
    """int: day of month or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[2]

  @property
  def hours(self):
    """int: number of hours or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[3]

  @property
  def minutes(self):
    """int: number of minutes or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[4]

  @property
  def month(self):
    """int: month or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[1]

  @property
  def seconds(self):
    """int: number of seconds or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[5]

  @property
  def year(self):
    """int: year or None if not set."""
    if not self._time_elements_tuple:
      return None
    return self._time_elements_tuple[0]

  def CopyFromDatetime(self, datetime_object):
    """Copies time elements from a Python datetime object.

    A naive datetime object is considered in local time.

    Args:
      datetime_object (datetime.datetime): Python datetime object.
    """
    year, month, day_of_month, hours, minutes, seconds, _, _, _ = (
        datetime_object.utctimetuple())

    date_time_values = {
        'year': year,
        'month': month,
        'day_of_month': day_of_month,
        'hours': hours,
        'minutes': minutes,
        'seconds': seconds}

    self._CopyFromDateTimeValues(date_time_values)

    self.is_local_time = bool(datetime_object.tzinfo is None)

  def CopyFromDateTimeString(self, time_string):
    """Copies time elements from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    self._CopyFromDateTimeValues(date_time_values)

  def CopyFromStringISO8601(self, time_string):
    """Copies time elements from an ISO 8601 date and time string.

    Currently not supported:
    * Duration notation: "P..."
    * Week notation "2016-W33"
    * Date with week number notation "2016-W33-3"
    * Date without year notation "--08-17"
    * Ordinal date notation "2016-230"

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DDThh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromStringISO8601(time_string)

    self._CopyFromDateTimeValues(date_time_values)

  def CopyFromStringRFC822(self, time_string):
    """Copies time elements from a RFC 822 date and time string.

    Args:
      time_string (str): date and time value formatted as:
          DAY, D MONTH YY hh:mm:ss ZONE

          Where weekday (DAY) and seconds (ss) are optional and day of
          month (D) can consist of 1 or 2 digits.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromStringRFC822(time_string)

    self._CopyFromDateTimeValues(date_time_values)

  def CopyFromStringRFC1123(self, time_string):
    """Copies time elements from a RFC 1123 date and time string.

    Args:
      time_string (str): date and time value formatted as:
          DAY, D MONTH YYYY hh:mm:ss ZONE

          Where weekday (DAY) and seconds (ss) are optional and day of
          month (D) can consist of 1 or 2 digits.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromStringRFC1123(time_string)

    self._CopyFromDateTimeValues(date_time_values)

  def CopyFromStringTuple(self, time_elements_tuple):
    """Copies time elements from string-based time elements tuple.

    Args:
      time_elements_tuple (Optional[tuple[str, str, str, str, str, str]]):
          time elements, contains year, month, day of month, hours, minutes and
          seconds.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    number_of_elements = len(time_elements_tuple)
    if number_of_elements < 6:
      raise ValueError((
          f'Invalid time elements tuple at least 6 elements required,'
          f'got: {number_of_elements:d}'))

    year_string = time_elements_tuple[0]
    month_string = time_elements_tuple[1]
    day_of_month_string = time_elements_tuple[2]
    hours_string = time_elements_tuple[3]
    minutes_string = time_elements_tuple[4]
    seconds_string = time_elements_tuple[5]

    try:
      year = int(year_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid year value: {year_string!s}')

    try:
      month = int(month_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid month value: {month_string!s}')

    try:
      day_of_month = int(day_of_month_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid day of month value: {day_of_month_string!s}')

    try:
      hours = int(hours_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid hours value: {hours_string!s}')

    try:
      minutes = int(minutes_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid minutes value: {minutes_string!s}')

    try:
      seconds = int(seconds_string, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid seconds value: {seconds_string!s}')

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds)

  def CopyToDateTimeString(self):
    """Copies the time elements to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss" or None
          if time elements are missing.
    """
    if self._number_of_seconds is None:
      return None

    year, month, day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    return (
        f'{year:04d}-{month:02d}-{day_of_month:02d} '
        f'{hours:02d}:{minutes:02d}:{seconds:02d}')

  def NewFromDeltaAndDate(self, year, month, day_of_month):
    """Creates a new time elements instance from a date time delta and a date.

    Args:
      year (int): year.
      month (int): month, where 1 represents January and 0 if not set.
      day_of_month (int): day of month, where 1 represents the first day and 0
          if not set.

    Returns:
      TimeElements: time elements or None if time elements are missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    if not self._is_delta:
      raise ValueError('Not a date time delta.')

    if self._time_elements_tuple is None:
      return None

    delta_year, delta_month, delta_day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    time_elements_tuple = (
        year + delta_year, month + delta_month,
        day_of_month + delta_day_of_month, hours, minutes, seconds)

    date_time = TimeElements(
        precision=self._precision, time_elements_tuple=time_elements_tuple,
        time_zone_offset=self._time_zone_offset)

    date_time.is_local_time = self.is_local_time

    return date_time

  def NewFromDeltaAndYear(self, year):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.

    Returns:
      TimeElements: time elements or None if time elements are missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    return self.NewFromDeltaAndDate(year, 0, 0)


class TimeElementsWithFractionOfSecond(TimeElements):
  """Time elements with a fraction of second interface.

  Attributes:
    fraction_of_second (decimal.Decimal): fraction of second, which must be a
        value between 0.0 and 1.0.
    is_local_time (bool): True if the date and time value is in local time.
  """

  def __init__(
      self, fraction_of_second=None, is_delta=False, precision=None,
      time_elements_tuple=None, time_zone_offset=None):
    """Initializes time elements.

    Args:
      fraction_of_second (Optional[decimal.Decimal]): fraction of second, which
          must be a value between 0.0 and 1.0.
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_elements_tuple (Optional[tuple[int, int, int, int, int, int]]):
          time elements, contains year, month, day of month, hours, minutes and
          seconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the time elements tuple is invalid or fraction of second
          value is out of bounds.
    """
    if fraction_of_second is not None:
      if fraction_of_second < 0.0 or fraction_of_second >= 1.0:
        raise ValueError(
            f'Fraction of second value: {fraction_of_second:f} out of bounds.')

    super(TimeElementsWithFractionOfSecond, self).__init__(
        is_delta=is_delta,
        precision=precision or definitions.PRECISION_1_SECOND,
        time_elements_tuple=time_elements_tuple,
        time_zone_offset=time_zone_offset)
    self.fraction_of_second = fraction_of_second

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if (self._number_of_seconds is not None and
          self.fraction_of_second is not None):
        self._normalized_timestamp = (
            decimal.Decimal(self._number_of_seconds) + self.fraction_of_second)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def _CopyFromDateTimeValues(self, date_time_values):
    """Copies time elements from date and time values.

    Args:
      date_time_values  (dict[str, int]): date and time values, such as year,
          month, day of month, hours, minutes, seconds, nanoseconds, time zone
          offset in minutes.

    Raises:
      ValueError: if no helper can be created for the current precision.
    """
    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    precision_helper = precisions.PrecisionHelperFactory.CreatePrecisionHelper(
        self._precision)

    fraction_of_second = precision_helper.CopyNanosecondsToFractionOfSecond(
        nanoseconds)

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds)
    self._time_zone_offset = time_zone_offset

    self.fraction_of_second = fraction_of_second

  def CopyFromDatetime(self, datetime_object):
    """Copies time elements from a Python datetime object.

    A naive datetime object is considered in local time.

    Args:
      datetime_object (datetime.datetime): Python datetime object.
    """
    super(TimeElementsWithFractionOfSecond, self).CopyFromDatetime(
        datetime_object)

    precision_helper = precisions.PrecisionHelperFactory.CreatePrecisionHelper(
        self._precision)

    fraction_of_second = precision_helper.CopyNanosecondsToFractionOfSecond(
        datetime_object.microsecond * definitions.NANOSECONDS_PER_MICROSECOND)
    self.fraction_of_second = fraction_of_second

  def CopyFromStringTuple(self, time_elements_tuple):
    """Copies time elements from string-based time elements tuple.

    Args:
      time_elements_tuple (Optional[tuple[str, str, str, str, str, str, str]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and fraction of seconds.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    number_of_elements = len(time_elements_tuple)
    if number_of_elements < 7:
      raise ValueError((
          f'Invalid time elements tuple at least 7 elements required,'
          f'got: {number_of_elements:d}'))

    super(TimeElementsWithFractionOfSecond, self).CopyFromStringTuple(
        time_elements_tuple)

    fraction_of_second_string = time_elements_tuple[6]

    try:
      fraction_of_second = decimal.Decimal(fraction_of_second_string)
    except (TypeError, ValueError):
      raise ValueError(
          f'Invalid fraction of second value: {fraction_of_second_string!s}')

    if fraction_of_second < 0.0 or fraction_of_second >= 1.0:
      raise ValueError(
          f'Fraction of second value: {fraction_of_second:f} out of bounds.')

    self.fraction_of_second = fraction_of_second

  def CopyToDateTimeString(self):
    """Copies the time elements to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss" or
          "YYYY-MM-DD hh:mm:ss.######" or None if time elements are missing.

    Raises:
      ValueError: if the precision value is unsupported.
    """
    if self._number_of_seconds is None or self.fraction_of_second is None:
      return None

    precision_helper = precisions.PrecisionHelperFactory.CreatePrecisionHelper(
        self._precision)

    return precision_helper.CopyToDateTimeString(
        self._time_elements_tuple, self.fraction_of_second)

  def NewFromDeltaAndDate(self, year, month, day_of_month):
    """Creates a new time elements instance from a date time delta and a date.

    Args:
      year (int): year.
      month (int): month, where 1 represents January and 0 if not set.
      day_of_month (int): day of month, where 1 represents the first day and 0
          if not set.

    Returns:
      TimeElementsWithFractionOfSecond: time elements or None if time elements
          are missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    if not self._is_delta:
      raise ValueError('Not a date time delta.')

    if self._time_elements_tuple is None:
      return None

    delta_year, delta_month, delta_day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    time_elements_tuple = (
        year + delta_year, month + delta_month,
        day_of_month + delta_day_of_month, hours, minutes, seconds)

    return TimeElementsWithFractionOfSecond(
        fraction_of_second=self.fraction_of_second, precision=self._precision,
        time_elements_tuple=time_elements_tuple,
        time_zone_offset=self._time_zone_offset)

  def NewFromDeltaAndYear(self, year):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.

    Returns:
      TimeElementsWithFractionOfSecond: time elements or None if time elements
          are missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    return self.NewFromDeltaAndDate(year, 0, 0)


class TimeElementsInMilliseconds(TimeElementsWithFractionOfSecond):
  """Time elements in milliseconds.

  Attributes:
    fraction_of_second (decimal.Decimal): fraction of second, which must be a
        value between 0.0 and 1.0.
    is_local_time (bool): True if the date and time value is in local time.
    precision (str): precision of the date of the date and time value, that
        represents 1 millisecond (PRECISION_1_MILLISECOND).
  """

  def __init__(
      self, is_delta=False, precision=None, time_elements_tuple=None,
      time_zone_offset=None):
    """Initializes time elements.

    Args:
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_elements_tuple (Optional[tuple[int, int, int, int, int, int, int]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and milliseconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    fraction_of_second = None
    if time_elements_tuple:
      number_of_elements = len(time_elements_tuple)
      if number_of_elements < 7:
        raise ValueError((
            f'Invalid time elements tuple at least 7 elements required,'
            f'got: {number_of_elements:d}'))

      milliseconds = time_elements_tuple[6]
      time_elements_tuple = time_elements_tuple[:6]

      if (milliseconds < 0 or
          milliseconds >= definitions.MILLISECONDS_PER_SECOND):
        raise ValueError('Invalid number of milliseconds.')

      fraction_of_second = (
          decimal.Decimal(milliseconds) / definitions.MILLISECONDS_PER_SECOND)

    super(TimeElementsInMilliseconds, self).__init__(
        fraction_of_second=fraction_of_second, is_delta=is_delta,
        precision=precision or definitions.PRECISION_1_MILLISECOND,
        time_elements_tuple=time_elements_tuple,
        time_zone_offset=time_zone_offset)

  @property
  def milliseconds(self):
    """int: number of milliseconds."""
    return int(self.fraction_of_second * definitions.MILLISECONDS_PER_SECOND)

  def CopyFromStringTuple(self, time_elements_tuple):
    """Copies time elements from string-based time elements tuple.

    Args:
      time_elements_tuple (Optional[tuple[str, str, str, str, str, str, str]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and milliseconds.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    number_of_elements = len(time_elements_tuple)
    if len(time_elements_tuple) < 7:
      raise ValueError((
          f'Invalid time elements tuple at least 7 elements required,'
          f'got: {number_of_elements:d}'))

    year, month, day_of_month, hours, minutes, seconds, milliseconds = (
        time_elements_tuple)

    try:
      milliseconds = int(milliseconds, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid millisecond value: {milliseconds!s}')

    if milliseconds < 0 or milliseconds >= definitions.MILLISECONDS_PER_SECOND:
      raise ValueError('Invalid number of milliseconds.')

    fraction_of_second = (
        decimal.Decimal(milliseconds) / definitions.MILLISECONDS_PER_SECOND)

    time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds,
        str(fraction_of_second))

    super(TimeElementsInMilliseconds, self).CopyFromStringTuple(
        time_elements_tuple)

  def NewFromDeltaAndDate(self, year, month, day_of_month):
    """Creates a new time elements instance from a date time delta and a date.

    Args:
      year (int): year.
      month (int): month, where 1 represents January and 0 if not set.
      day_of_month (int): day of month, where 1 represents the first day and 0
          if not set.

    Returns:
      TimeElementsInMilliseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    if not self._is_delta:
      raise ValueError('Not a date time delta.')

    if self._time_elements_tuple is None:
      return None

    delta_year, delta_month, delta_day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    time_elements_tuple = (
        year + delta_year, month + delta_month,
        day_of_month + delta_day_of_month, hours, minutes, seconds,
        self.milliseconds)

    return TimeElementsInMilliseconds(
        precision=self._precision, time_elements_tuple=time_elements_tuple,
        time_zone_offset=self._time_zone_offset)

  def NewFromDeltaAndYear(self, year):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.

    Returns:
      TimeElementsInMilliseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    return self.NewFromDeltaAndDate(year, 0, 0)


class TimeElementsInMicroseconds(TimeElementsWithFractionOfSecond):
  """Time elements in microseconds.

  Attributes:
    fraction_of_second (decimal.Decimal): fraction of second, which must be a
        value between 0.0 and 1.0.
    is_local_time (bool): True if the date and time value is in local time.
    precision (str): precision of the date of the date and time value, that
        represents 1 microsecond (PRECISION_1_MICROSECOND).
  """

  def __init__(
      self, is_delta=False, precision=None, time_elements_tuple=None,
      time_zone_offset=None):
    """Initializes time elements.

    Args:
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_elements_tuple (Optional[tuple[int, int, int, int, int, int, int]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and microseconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    fraction_of_second = None
    if time_elements_tuple:
      number_of_elements = len(time_elements_tuple)
      if number_of_elements < 7:
        raise ValueError((
            f'Invalid time elements tuple at least 7 elements required,'
            f'got: {number_of_elements:d}'))

      microseconds = time_elements_tuple[6]
      time_elements_tuple = time_elements_tuple[:6]

      if (microseconds < 0 or
          microseconds >= definitions.MICROSECONDS_PER_SECOND):
        raise ValueError('Invalid number of microseconds.')

      fraction_of_second = (
          decimal.Decimal(microseconds) / definitions.MICROSECONDS_PER_SECOND)

    super(TimeElementsInMicroseconds, self).__init__(
        fraction_of_second=fraction_of_second, is_delta=is_delta,
        precision=precision or definitions.PRECISION_1_MICROSECOND,
        time_elements_tuple=time_elements_tuple,
        time_zone_offset=time_zone_offset)

  @property
  def microseconds(self):
    """int: number of microseconds."""
    return int(self.fraction_of_second * definitions.MICROSECONDS_PER_SECOND)

  def CopyFromStringTuple(self, time_elements_tuple):
    """Copies time elements from string-based time elements tuple.

    Args:
      time_elements_tuple (Optional[tuple[str, str, str, str, str, str, str]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and microseconds.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    number_of_elements = len(time_elements_tuple)
    if len(time_elements_tuple) < 7:
      raise ValueError((
          f'Invalid time elements tuple at least 7 elements required,'
          f'got: {number_of_elements:d}'))

    year, month, day_of_month, hours, minutes, seconds, microseconds = (
        time_elements_tuple)

    try:
      microseconds = int(microseconds, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid microsecond value: {microseconds!s}')

    if microseconds < 0 or microseconds >= definitions.MICROSECONDS_PER_SECOND:
      raise ValueError('Invalid number of microseconds.')

    fraction_of_second = (
        decimal.Decimal(microseconds) / definitions.MICROSECONDS_PER_SECOND)

    time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds,
        str(fraction_of_second))

    super(TimeElementsInMicroseconds, self).CopyFromStringTuple(
        time_elements_tuple)

  def NewFromDeltaAndDate(self, year, month, day_of_month):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.
      month (int): month, where 1 represents January and 0 if not set.
      day_of_month (int): day of month, where 1 represents the first day and 0
          if not set.

    Returns:
      TimeElementsInMicroseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    if not self._is_delta:
      raise ValueError('Not a date time delta.')

    if self._time_elements_tuple is None:
      return None

    delta_year, delta_month, delta_day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    time_elements_tuple = (
        year + delta_year, month + delta_month,
        day_of_month + delta_day_of_month, hours, minutes, seconds,
        self.microseconds)

    return TimeElementsInMicroseconds(
        precision=self._precision, time_elements_tuple=time_elements_tuple,
        time_zone_offset=self._time_zone_offset)

  def NewFromDeltaAndYear(self, year):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.

    Returns:
      TimeElementsInMicroseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    return self.NewFromDeltaAndDate(year, 0, 0)


class TimeElementsInNanoseconds(TimeElementsWithFractionOfSecond):
  """Time elements in nanoseconds.

  Attributes:
    fraction_of_second (decimal.Decimal): fraction of second, which must be a
        value between 0.0 and 1.0.
    is_local_time (bool): True if the date and time value is in local time.
    precision (str): precision of the date of the date and time value, that
        represents 1 nanosecond (PRECISION_1_NANOSECOND).
  """

  def __init__(
      self, is_delta=False, precision=None, time_elements_tuple=None,
      time_zone_offset=None):
    """Initializes time elements.

    Args:
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_elements_tuple (Optional[tuple[int, int, int, int, int, int, int]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and nanoseconds.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    fraction_of_second = None
    if time_elements_tuple:
      number_of_elements = len(time_elements_tuple)
      if number_of_elements < 7:
        raise ValueError((
            f'Invalid time elements tuple at least 7 elements required,'
            f'got: {number_of_elements:d}'))

      nanoseconds = time_elements_tuple[6]
      time_elements_tuple = time_elements_tuple[:6]

      if (nanoseconds < 0 or
          nanoseconds >= definitions.NANOSECONDS_PER_SECOND):
        raise ValueError('Invalid number of nanoseconds.')

      fraction_of_second = (
          decimal.Decimal(nanoseconds) / definitions.NANOSECONDS_PER_SECOND)

    super(TimeElementsInNanoseconds, self).__init__(
        fraction_of_second=fraction_of_second, is_delta=is_delta,
        precision=precision or definitions.PRECISION_1_NANOSECOND,
        time_elements_tuple=time_elements_tuple,
        time_zone_offset=time_zone_offset)

  @property
  def nanoseconds(self):
    """int: number of nanoseconds."""
    return int(self.fraction_of_second * definitions.NANOSECONDS_PER_SECOND)

  def CopyFromStringTuple(self, time_elements_tuple):
    """Copies time elements from string-based time elements tuple.

    Args:
      time_elements_tuple (Optional[tuple[str, str, str, str, str, str, str]]):
          time elements, contains year, month, day of month, hours, minutes,
          seconds and nanoseconds.

    Raises:
      ValueError: if the time elements tuple is invalid.
    """
    number_of_elements = len(time_elements_tuple)
    if len(time_elements_tuple) < 7:
      raise ValueError((
          f'Invalid time elements tuple at least 7 elements required,'
          f'got: {number_of_elements:d}'))

    year, month, day_of_month, hours, minutes, seconds, nanoseconds = (
        time_elements_tuple)

    try:
      nanoseconds = int(nanoseconds, 10)
    except (TypeError, ValueError):
      raise ValueError(f'Invalid nanosecond value: {nanoseconds!s}')

    if nanoseconds < 0 or nanoseconds >= definitions.NANOSECONDS_PER_SECOND:
      raise ValueError('Invalid number of nanoseconds.')

    fraction_of_second = (
        decimal.Decimal(nanoseconds) / definitions.NANOSECONDS_PER_SECOND)

    time_elements_tuple = (
        year, month, day_of_month, hours, minutes, seconds,
        str(fraction_of_second))

    super(TimeElementsInNanoseconds, self).CopyFromStringTuple(
        time_elements_tuple)

  def NewFromDeltaAndDate(self, year, month, day_of_month):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.
      month (int): month, where 1 represents January and 0 if not set.
      day_of_month (int): day of month, where 1 represents the first day and 0
          if not set.

    Returns:
      TimeElementsInNanoseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    if not self._is_delta:
      raise ValueError('Not a date time delta.')

    if self._time_elements_tuple is None:
      return None

    delta_year, delta_month, delta_day_of_month, hours, minutes, seconds = (
        self._time_elements_tuple)

    time_elements_tuple = (
        year + delta_year, month + delta_month,
        day_of_month + delta_day_of_month, hours, minutes, seconds,
        self.nanoseconds)

    return TimeElementsInNanoseconds(
        precision=self._precision, time_elements_tuple=time_elements_tuple,
        time_zone_offset=self._time_zone_offset)

  def NewFromDeltaAndYear(self, year):
    """Creates a new time elements instance from a date time delta and a year.

    Args:
      year (int): year.

    Returns:
      TimeElementsInNanoseconds: time elements or None if time elements are
          missing.

    Raises:
      ValueError: if the instance is not a date time delta.
    """
    return self.NewFromDeltaAndDate(year, 0, 0)


factory.Factory.RegisterDateTimeValues(TimeElements)
factory.Factory.RegisterDateTimeValues(TimeElementsInMilliseconds)
factory.Factory.RegisterDateTimeValues(TimeElementsInMicroseconds)
factory.Factory.RegisterDateTimeValues(TimeElementsInNanoseconds)



# -*- coding: utf-8 -*-
"""Golang time.Time timestamp implementation."""

import decimal
import struct

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import interface


class GolangTimeEpoch(interface.DateTimeEpoch):
  """Golang time.Time epoch."""

  def __init__(self):
    """Initializes a Golang time.Time epoch."""
    super(GolangTimeEpoch, self).__init__(1, 1, 1)


class GolangTime(interface.DateTimeValues):
  """Golang time.Time timestamp.

  A Golang time.Time timestamp contains the number of nanoseconds since
  January 1, 1 UTC. Depending on the version of the timestamp, the time
  zone is stored in minutes or seconds relative to UTC.

  A serialized version 1 Golang time.Time timestamp is a 15 byte value
  that consists of:

  * byte 0 - version as an 8-bit integer.
  * bytes 1-8 - number of seconds since January 1, 1 as a big-endian signed
      integer.
  * bytes 9-12 - fraction of second, number of nanoseconds as a big-endian
      signed integer.
  * bytes 13-14 - time zone offset in minutes as a 16-bit big-endian integer,
      where -1 represents UTC.

  A serialized version 2 Golang time.Time timestamp is a 16 byte value
  that consists of:

  * byte 0 - version as an 8-bit integer.
  * bytes 1-8 - number of seconds since January 1, 1 as a big-endian signed
      integer.
  * bytes 9-12 - fraction of second, number of nanoseconds as a big-endian
      signed integer.
  * bytes 13-14 - time zone offset in minutes as a 16-bit big-endian integer,
      where -1 represents UTC.
  * byte 15 - time zone offset in seconds as an 8-bit integer.

  Attributes:
    is_local_time (bool): True if the date and time value is in local time
  """

  # The delta between January 1, 1970 (unix epoch) and January 1, 1
  # (Golang epoch).
  _GOLANG_TO_POSIX_BASE = (
      ((1969 * 365) + (1969 // 4) - (1969 // 100) + (1969 // 400)) *
      definitions.SECONDS_PER_DAY)

  _EPOCH = GolangTimeEpoch()

  def __init__(self, golang_timestamp=None, precision=None):
    """Initializes a Golang time.Time timestamp.

    Args:
      golang_timestamp (Optional[bytes]): the Golang time.Time timestamp.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
    """
    number_of_seconds, nanoseconds, time_zone_offset = (None, None, None)
    if golang_timestamp is not None:
      number_of_seconds, nanoseconds, time_zone_offset = (
          self._GetNumberOfSeconds(golang_timestamp))

    super(GolangTime, self).__init__(
        precision=precision or definitions.PRECISION_1_NANOSECOND,
        time_zone_offset=time_zone_offset)
    self._golang_timestamp = golang_timestamp
    self._nanoseconds = nanoseconds
    self._number_of_seconds = number_of_seconds

  @property
  def golang_timestamp(self):
    """int: Golang time.Time timestamp or None if not set."""
    return self._golang_timestamp

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if (self._number_of_seconds is not None and
          self._number_of_seconds >= self._GOLANG_TO_POSIX_BASE and
          self._nanoseconds is not None and self._nanoseconds >= 0):

        self._normalized_timestamp = decimal.Decimal(
            self._number_of_seconds - GolangTime._GOLANG_TO_POSIX_BASE)

        if self._nanoseconds is not None and self._nanoseconds >= 0:
          self._normalized_timestamp += (
              decimal.Decimal(self._nanoseconds) /
              definitions.NANOSECONDS_PER_SECOND)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def _GetNumberOfSeconds(self, golang_timestamp):
    """Retrieves the number of seconds from a Golang time.Time timestamp.

    Args:
      golang_timestamp (bytes): the Golang time.Time timestamp.

    Returns:
      tuple[int, int, int]: number of seconds since January 1, 1 00:00:00,
          fraction of second in nanoseconds and time zone offset in minutes.

    Raises:
      ValueError: if the Golang time.Time timestamp could not be parsed.
    """
    byte_size = len(golang_timestamp)
    if byte_size < 15:
      raise ValueError('Unsupported Golang time.Time timestamp.')

    version = golang_timestamp[0]
    if version not in (1, 2):
      raise ValueError(
          f'Unsupported Golang time.Time timestamp version: {version:d}.')

    if (version == 1 and byte_size != 15) or (version == 2 and byte_size != 16):
      raise ValueError('Unsupported Golang time.Time timestamp.')

    try:
      number_of_seconds, nanoseconds, time_zone_offset = struct.unpack(
          '>qih', golang_timestamp[1:15])

      # TODO: add support for version 2 time zone offset in seconds

    except struct.error as exception:
      raise ValueError((
          f'Unable to unpacked Golang time.Time timestamp with error: '
          f'{exception!s}'))

    # A time zone offset of -1 minute is a special representation for UTC.
    if time_zone_offset == -1:
      time_zone_offset = 0

    return number_of_seconds, nanoseconds, time_zone_offset

  def CopyFromDateTimeString(self, time_string):
    """Copies a date time value from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)
    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    if year < 0:
      raise ValueError(f'Year value not supported: {year!s}.')

    seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)

    seconds += self._GOLANG_TO_POSIX_BASE

    self._normalized_timestamp = None
    self._number_of_seconds = seconds
    self._nanoseconds = nanoseconds
    self._time_zone_offset = time_zone_offset

  def CopyToDateTimeString(self):
    """Copies the Golang time value to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.######" or
          None if the timestamp cannot be copied to a date and time string.
    """
    if self._number_of_seconds is None or self._number_of_seconds < 0:
      return None

    number_of_days, hours, minutes, seconds = self._GetTimeValues(
        self._number_of_seconds)

    year, month, day_of_month = self._GetDateValuesWithEpoch(
        number_of_days, self._EPOCH)

    return (f'{year:04d}-{month:02d}-{day_of_month:02d} '
            f'{hours:02d}:{minutes:02d}:{seconds:02d}.{self._nanoseconds:09d}')


factory.Factory.RegisterDateTimeValues(GolangTime)



# -*- coding: utf-8 -*-
"""Fake timestamp implementation."""

import decimal
import time

from dfdatetime import definitions
from dfdatetime import interface
from dfdatetime import posix_time


class FakeTime(interface.DateTimeValues):
  """Fake timestamp.

  The fake timestamp is intended for testing purposes. On initialization
  it contains the current time in UTC in microsecond precision.

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
  """

  _EPOCH = posix_time.PosixTimeEpoch()

  def __init__(self, precision=None, time_zone_offset=None):
    """Initializes a fake timestamp.

    Args:
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.
    """
    # Note that time.time() and divmod return floating point values.
    timestamp, fraction_of_second = divmod(time.time(), 1)

    super(FakeTime, self).__init__(
        precision=precision or definitions.PRECISION_1_MICROSECOND,
        time_zone_offset=time_zone_offset)
    self._microseconds = int(
        fraction_of_second * definitions.MICROSECONDS_PER_SECOND)
    self._number_of_seconds = int(timestamp)

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._number_of_seconds is not None:
        self._normalized_timestamp = (
            decimal.Decimal(self._microseconds) /
            definitions.MICROSECONDS_PER_SECOND)
        self._normalized_timestamp += decimal.Decimal(self._number_of_seconds)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def CopyFromDateTimeString(self, time_string):
    """Copies a fake timestamp from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', None)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)

    if nanoseconds is None:
      self._microseconds = None
    else:
      self._microseconds, _ = divmod(nanoseconds, 1000)

    self._time_zone_offset = time_zone_offset

  def CopyToDateTimeString(self):
    """Copies the fake timestamp to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss" or
          "YYYY-MM-DD hh:mm:ss.######" or None if the number of seconds
          is missing.
    """
    if self._number_of_seconds is None:
      return None

    number_of_days, hours, minutes, seconds = self._GetTimeValues(
        self._number_of_seconds)

    year, month, day_of_month = self._GetDateValuesWithEpoch(
        number_of_days, self._EPOCH)

    date_time_string = (
        f'{year:04d}-{month:02d}-{day_of_month:02d} '
        f'{hours:02d}:{minutes:02d}:{seconds:02d}')

    if self._microseconds is not None:
      date_time_string = '.'.join([
          date_time_string, f'{self._microseconds:06d}'])

    return date_time_string



# -*- coding: utf-8 -*-
"""Java java.util.Date timestamp implementation."""

import decimal

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import posix_time


class JavaTime(posix_time.PosixTimeInMilliseconds):
  """Java java.util.Date timestamp.

  The Java java.util.Date timestamp is a signed integer that contains the
  number of milliseconds since 1970-01-01 00:00:00 (also known as the POSIX
  epoch). Negative values represent date and times predating the POSIX epoch.

  Also see:
    https://docs.oracle.com/javase/8/docs/api/java/util/Date.html

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
  """

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if (self._timestamp is not None and self._timestamp >= self._INT64_MIN and
          self._timestamp <= self._INT64_MAX):
        self._normalized_timestamp = (
            decimal.Decimal(self._timestamp) /
            definitions.MILLISECONDS_PER_SECOND)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def CopyToDateTimeString(self):
    """Copies the POSIX timestamp to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.######" or
          None if the timestamp is missing.
    """
    if (self._timestamp is None or self._timestamp < self._INT64_MIN or
        self._timestamp > self._INT64_MAX):
      return None

    return super(JavaTime, self).CopyToDateTimeString()


factory.Factory.RegisterDateTimeValues(JavaTime)



# -*- coding: utf-8 -*-
"""FAT date time implementation."""

import decimal

from dfdatetime import definitions
from dfdatetime import factory
from dfdatetime import interface


class FATDateTimeEpoch(interface.DateTimeEpoch):
  """FAT date time time epoch."""

  def __init__(self):
    """Initializes a FAT date time epoch."""
    super(FATDateTimeEpoch, self).__init__(1980, 1, 1)


class FATDateTime(interface.DateTimeValues):
  """FAT date time.

  The FAT date time is mainly used in DOS/Windows file formats and FAT.

  The FAT date and time is a 32-bit value containing two 16-bit values:
    * The date (lower 16-bit).
      * bits 0 - 4: day of month, where 1 represents the first day
      * bits 5 - 8: month of year, where 1 represent January
      * bits 9 - 15: year since 1980
    * The time of day (upper 16-bit).
      * bits 0 - 4: seconds (in 2 second intervals)
      * bits 5 - 10: minutes
      * bits 11 - 15: hours

  The FAT date time has no time zone information and is typically stored
  in the local time of the computer.

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
  """

  _EPOCH = FATDateTimeEpoch()

  # The difference between January 1, 1980 and January 1, 1970 in seconds.
  _FAT_DATE_TO_POSIX_BASE = 315532800

  def __init__(self, fat_date_time=None, precision=None, time_zone_offset=None):
    """Initializes a FAT date time.

    Args:
      fat_date_time (Optional[int]): FAT date time.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.
    """
    super(FATDateTime, self).__init__(
        precision=precision or definitions.PRECISION_2_SECONDS,
        time_zone_offset=time_zone_offset)
    self._fat_date_time = fat_date_time
    self._number_of_seconds = None

    if fat_date_time is not None:
      self._number_of_seconds = self._GetNumberOfSeconds(fat_date_time)

  @property
  def fat_date_time(self):
    """int: FAT date time or None if not set."""
    return self._fat_date_time

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._number_of_seconds is not None and self._number_of_seconds >= 0:
        self._normalized_timestamp = (
            decimal.Decimal(self._number_of_seconds) +
            self._FAT_DATE_TO_POSIX_BASE)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def _GetNumberOfSeconds(self, fat_date_time):
    """Retrieves the number of seconds from a FAT date time.

    Args:
      fat_date_time (int): FAT date time.

    Returns:
      int: number of seconds since January 1, 1980 00:00:00.

    Raises:
      ValueError: if the month, day of month, hours, minutes or seconds
          value is out of bounds.
    """
    day_of_month = fat_date_time & 0x1f
    month = (fat_date_time >> 5) & 0x0f
    year = (fat_date_time >> 9) & 0x7f

    days_per_month = self._GetDaysPerMonth(year, month)
    if day_of_month < 1 or day_of_month > days_per_month:
      raise ValueError('Day of month value out of bounds.')

    number_of_days = self._GetDayOfYear(1980 + year, month, day_of_month)
    number_of_days -= 1
    for past_year in range(0, year):
      number_of_days += self._GetNumberOfDaysInYear(past_year)

    fat_date_time >>= 16

    seconds = (fat_date_time & 0x1f) * 2
    minutes = (fat_date_time >> 5) & 0x3f
    hours = (fat_date_time >> 11) & 0x1f

    if hours not in range(0, 24):
      raise ValueError('Hours value out of bounds.')

    if minutes not in range(0, 60):
      raise ValueError('Minutes value out of bounds.')

    if seconds not in range(0, 60):
      raise ValueError('Seconds value out of bounds.')

    number_of_seconds = (((hours * 60) + minutes) * 60) + seconds
    number_of_seconds += number_of_days * definitions.SECONDS_PER_DAY
    return number_of_seconds

  def CopyFromDateTimeString(self, time_string):
    """Copies a FAT date time from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    if year < 1980 or year > (1980 + 0x7f):
      raise ValueError(f'Year value not supported: {year!s}.')

    self._normalized_timestamp = None
    self._number_of_seconds = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    self._number_of_seconds -= self._FAT_DATE_TO_POSIX_BASE
    self._time_zone_offset = time_zone_offset

  def CopyToDateTimeString(self):
    """Copies the FAT date time to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss" or None
          if number of seconds is missing.
    """
    if self._number_of_seconds is None:
      return None

    number_of_days, hours, minutes, seconds = self._GetTimeValues(
        self._number_of_seconds)

    year, month, day_of_month = self._GetDateValuesWithEpoch(
        number_of_days, self._EPOCH)

    return (f'{year:04d}-{month:02d}-{day_of_month:02d} '
            f'{hours:02d}:{minutes:02d}:{seconds:02d}')


class FATTimestamp(interface.DateTimeValues):
  """FAT timestamp.

  The FAT timestamp is an unsigned integer that contains the number of
  10 milli seconds intervals since 1980-01-01 00:00:00 (also known as
  the FAT date time epoch).

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
  """

  _EPOCH = FATDateTimeEpoch()

  # The difference between January 1, 1980 and January 1, 1970 in seconds.
  _FAT_DATE_TO_POSIX_BASE = 315532800

  def __init__(self, precision=None, time_zone_offset=None, timestamp=None):
    """Initializes a FAT timestamp.

    Args:
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.
      timestamp (Optional[int]): FAT timestamp.
    """
    super(FATTimestamp, self).__init__(
        precision=precision or definitions.PRECISION_10_MILLISECONDS,
        time_zone_offset=time_zone_offset)
    self._timestamp = timestamp

  @property
  def timestamp(self):
    """int: FAT timestamp or None if not set."""
    return self._timestamp

  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """
    if self._normalized_timestamp is None:
      if self._timestamp is not None:
        self._normalized_timestamp = (
            (decimal.Decimal(self._timestamp) / 100) +
            self._FAT_DATE_TO_POSIX_BASE)

        if self._time_zone_offset:
          self._normalized_timestamp -= self._time_zone_offset * 60

    return self._normalized_timestamp

  def CopyFromDateTimeString(self, time_string):
    """Copies a FAT timestamp from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    date_time_values = self._CopyDateTimeFromString(time_string)

    year = date_time_values.get('year', 0)
    month = date_time_values.get('month', 0)
    day_of_month = date_time_values.get('day_of_month', 0)
    hours = date_time_values.get('hours', 0)
    minutes = date_time_values.get('minutes', 0)
    seconds = date_time_values.get('seconds', 0)
    nanoseconds = date_time_values.get('nanoseconds', 0)
    time_zone_offset = date_time_values.get('time_zone_offset', None)

    if year < 1980 or year > (1980 + 0x7f):
      raise ValueError(f'Year value not supported: {year!s}.')

    milliseconds, _ = divmod(nanoseconds, 10000000)

    timestamp = self._GetNumberOfSecondsFromElements(
        year, month, day_of_month, hours, minutes, seconds)
    timestamp -= self._FAT_DATE_TO_POSIX_BASE
    timestamp *= 100
    timestamp += milliseconds

    self._timestamp = timestamp
    self._time_zone_offset = time_zone_offset

  def CopyToDateTimeString(self):
    """Copies the FAT timestamp to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.######" or
          None if the timestamp is missing.
    """
    if self._timestamp is None:
      return None

    timestamp, milliseconds = divmod(self._timestamp, 100)
    number_of_days, hours, minutes, seconds = self._GetTimeValues(timestamp)

    year, month, day_of_month = self._GetDateValuesWithEpoch(
        number_of_days, self._EPOCH)

    return (f'{year:04d}-{month:02d}-{day_of_month:02d} '
            f'{hours:02d}:{minutes:02d}:{seconds:02d}.{milliseconds:02d}')


factory.Factory.RegisterDateTimeValues(FATDateTime)
factory.Factory.RegisterDateTimeValues(FATTimestamp)



# -*- coding: utf-8 -*-
"""Date and time interfaces."""

import abc
import decimal

from dfdatetime import definitions


class DateTimeEpoch(object):
  """Date and time epoch interface.

  This is the super class of different epoch representations.

  Attributes:
    year (int): year that is the start of the epoch e.g. 1970.
    month (int): month that is the start of the epoch, where 1 represents
        January.
    day_of_month (int): day of the month that is the start of the epoch,
        where 1 represents the first day.
  """

  def __init__(self, year, month, day_of_month):
    """Initializes a date time epoch.

    Args:
      year (int): year that is the start of the epoch e.g. 1970.
      month (int): month that is the start of the epoch, where 1 represents
          January.
      day_of_month (int): day of the month that is the start of the epoch,
          where 1 represents the first day.
    """
    super(DateTimeEpoch, self).__init__()
    self.day_of_month = day_of_month
    self.month = month
    self.year = year


class NormalizedTimeEpoch(DateTimeEpoch):
  """dfDateTime normalized time epoch."""

  def __init__(self):
    """Initializes a dfDateTime normalized time epoch."""
    super(NormalizedTimeEpoch, self).__init__(1970, 1, 1)


class DateTimeValues(object):
  """Date and time values interface.

  This is the super class of different date and time representations.

  Attributes:
    is_local_time (bool): True if the date and time value is in local time.
    time_zone_hint (str): time zone hint, such as "Europe/Amsterdam", "CET" or
        "UTC+1", or None if not set.
  """

  # pylint: disable=redundant-returns-doc

  _EPOCH_NORMALIZED_TIME = NormalizedTimeEpoch()

  _100_MILLISECONDS_PER_SECOND = 10
  _10_MILLISECONDS_PER_SECOND = 100
  _1_MILLISECOND_PER_SECOND = 1000
  _100_MICROSECONDS_PER_SECOND = 10000
  _10_MICROSECONDS_PER_SECOND = 100000
  _1_MICROSECOND_PER_SECOND = 1000000
  _100_NANOSECONDS_PER_SECOND = 10000000
  _10_NANOSECONDS_PER_SECOND = 100000000
  _1_NANOSECOND_PER_SECOND = definitions.NANOSECONDS_PER_SECOND

  _100_NANOSECONDS_PER_MICROSECOND = 10

  _INT64_MIN = -(1 << 63)
  _INT64_MAX = (1 << 63) - 1

  _UINT32_MAX = (1 << 32) - 1
  _UINT60_MAX = (1 << 60) - 1
  _UINT64_MAX = (1 << 64) - 1

  _REMAINDER_MULTIPLIER = {
      definitions.PRECISION_1_MILLISECOND: _1_MILLISECOND_PER_SECOND,
      definitions.PRECISION_10_MILLISECONDS: _10_MILLISECONDS_PER_SECOND,
      definitions.PRECISION_100_MILLISECONDS: _100_MILLISECONDS_PER_SECOND,
      definitions.PRECISION_1_MICROSECOND: _1_MICROSECOND_PER_SECOND,
      definitions.PRECISION_10_MICROSECONDS: _10_MICROSECONDS_PER_SECOND,
      definitions.PRECISION_100_MICROSECONDS: _100_MICROSECONDS_PER_SECOND,
      definitions.PRECISION_1_NANOSECOND: _1_NANOSECOND_PER_SECOND,
      definitions.PRECISION_10_NANOSECONDS: _10_NANOSECONDS_PER_SECOND,
      definitions.PRECISION_100_NANOSECONDS: _100_NANOSECONDS_PER_SECOND}

  def __init__(self, is_delta=False, precision=None, time_zone_offset=None):
    """Initializes date time values.

    Args:
      is_delta (Optional[bool]): True if the date and time value is relative to
          another date and time value.
      precision (Optional[str]): precision of the date and time value, which
          should be one of the PRECISION_VALUES in definitions.
      time_zone_offset (Optional[int]): time zone offset in number of minutes
          from UTC or None if not set.
    """
    super(DateTimeValues, self).__init__()
    self._cached_date_time_values = None
    self._is_delta = is_delta
    self._normalized_timestamp = None
    self._precision = precision
    self._time_zone_offset = time_zone_offset

    self.is_local_time = False
    self.time_zone_hint = False

  @property
  def is_delta(self):
    """Is delta (bool): True if the date and time is relative to another."""
    return self._is_delta

  @property
  def precision(self):
    """Precision (str): precision of the date and time value."""
    return self._precision

  @property
  def time_zone_offset(self):
    """Time zone offset (int): time zone offset in minutes from UTC."""
    return self._time_zone_offset

  @time_zone_offset.setter
  def time_zone_offset(self, time_zone_offset):
    """Sets the time zone offset.

    Args:
      time_zone_offset (int): time zone offset in number of minutes from UTC or
          None if not set.
    """
    self._normalized_timestamp = None
    self._time_zone_offset = time_zone_offset

  def __eq__(self, other):
    """Determines if the date time values are equal to other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are equal to other.
    """
    if not isinstance(other, DateTimeValues):
      return False

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None and other_normalized_timestamp is not None:
      return False

    if normalized_timestamp is not None and other_normalized_timestamp is None:
      return False

    return normalized_timestamp == other_normalized_timestamp

  def __ge__(self, other):
    """Determines if the date time values are greater than or equal to other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are greater than or equal to other.

    Raises:
      ValueError: if other is not an instance of DateTimeValues.
    """
    if not isinstance(other, DateTimeValues):
      raise ValueError('Other not an instance of DateTimeValues')

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None:
      return other_normalized_timestamp is None

    if other_normalized_timestamp is None:
      return True

    return normalized_timestamp >= other_normalized_timestamp

  def __gt__(self, other):
    """Determines if the date time values are greater than other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are greater than other.

    Raises:
      ValueError: if other is not an instance of DateTimeValues.
    """
    if not isinstance(other, DateTimeValues):
      raise ValueError('Other not an instance of DateTimeValues')

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None:
      return False

    if other_normalized_timestamp is None:
      return True

    return normalized_timestamp > other_normalized_timestamp

  def __le__(self, other):
    """Determines if the date time values are greater than or equal to other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are greater than or equal to other.

    Raises:
      ValueError: if other is not an instance of DateTimeValues.
    """
    if not isinstance(other, DateTimeValues):
      raise ValueError('Other not an instance of DateTimeValues')

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None:
      return True

    if other_normalized_timestamp is None:
      return False

    return normalized_timestamp <= other_normalized_timestamp

  def __lt__(self, other):
    """Determines if the date time values are less than other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are less than other.

    Raises:
      ValueError: if other is not an instance of DateTimeValues.
    """
    if not isinstance(other, DateTimeValues):
      raise ValueError('Other not an instance of DateTimeValues')

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None:
      return other_normalized_timestamp is not None

    if other_normalized_timestamp is None:
      return False

    return normalized_timestamp < other_normalized_timestamp

  def __ne__(self, other):
    """Determines if the date time values are not equal to other.

    Args:
      other (DateTimeValues): date time values to compare against.

    Returns:
      bool: True if the date time values are not equal to other.
    """
    if not isinstance(other, DateTimeValues):
      return True

    normalized_timestamp = self._GetNormalizedTimestamp()
    other_normalized_timestamp = other._GetNormalizedTimestamp()  # pylint: disable=protected-access

    if normalized_timestamp is None and other_normalized_timestamp is not None:
      return True

    if normalized_timestamp is not None and other_normalized_timestamp is None:
      return True

    return normalized_timestamp != other_normalized_timestamp

  def _CopyDateFromString(self, date_string):
    """Copies a date from a string.

    Args:
      date_string (str): date value formatted as: YYYY-MM-DD

    Returns:
      tuple[int, int, int]: year, month, day of month.

    Raises:
      ValueError: if the date string is invalid or not supported.
    """
    date_string_length = len(date_string)

    # The date string should at least contain 'YYYY-MM-DD'.
    if date_string_length < 10:
      raise ValueError('Date string too short.')

    if date_string[4] != '-' or date_string[7] != '-':
      raise ValueError('Invalid date string.')

    try:
      year = int(date_string[0:4], 10)
    except ValueError:
      raise ValueError('Unable to parse year.')

    try:
      month = int(date_string[5:7], 10)
    except ValueError:
      raise ValueError('Unable to parse month.')

    try:
      day_of_month = int(date_string[8:10], 10)
    except ValueError:
      raise ValueError('Unable to parse day of month.')

    days_per_month = self._GetDaysPerMonth(year, month)
    if day_of_month < 1 or day_of_month > days_per_month:
      raise ValueError('Day of month value out of bounds.')

    return year, month, day_of_month

  def _CopyDateTimeFromString(self, time_string):
    """Copies a date and time from a string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Returns:
      dict[str, int]: date and time values, such as year, month, day of month,
          hours, minutes, seconds, nanoseconds, time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    if not time_string:
      raise ValueError('Invalid time string.')

    time_string_length = len(time_string)

    year, month, day_of_month = self._CopyDateFromString(time_string)

    if time_string_length <= 10:
      return {
          'year': year,
          'month': month,
          'day_of_month': day_of_month}

    # If a time of day is specified the time string it should at least
    # contain 'YYYY-MM-DD hh:mm:ss'.
    if time_string[10] != ' ':
      raise ValueError(
          'Invalid time string - space missing as date and time separator.')

    hours, minutes, seconds, nanoseconds, time_zone_offset = (
        self._CopyTimeFromString(time_string[11:]))

    date_time_values = {
        'year': year,
        'month': month,
        'day_of_month': day_of_month,
        'hours': hours,
        'minutes': minutes,
        'seconds': seconds}

    if nanoseconds is not None:
      date_time_values['nanoseconds'] = nanoseconds
    if time_zone_offset is not None:
      date_time_values['time_zone_offset'] = time_zone_offset

    return date_time_values

  def _CopyTimeFromString(self, time_string):
    """Copies a time from a string.

    Args:
      time_string (str): time value formatted as:
          hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The seconds fraction and
          time zone offset are optional.

    Returns:
      tuple[int, int, int, int, int]: hours, minutes, seconds, nanoseconds,
          time zone offset in minutes.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """
    time_string_length = len(time_string)

    # The time string should at least contain 'hh:mm:ss'.
    if time_string_length < 8:
      raise ValueError('Time string too short.')

    if time_string[2] != ':' or time_string[5] != ':':
      raise ValueError('Invalid time string.')

    try:
      hours = int(time_string[0:2], 10)
    except ValueError:
      raise ValueError('Unable to parse hours.')

    if hours not in range(0, 24):
      raise ValueError(f'Hours value: {hours:d} out of bounds.')

    try:
      minutes = int(time_string[3:5], 10)
    except ValueError:
      raise ValueError('Unable to parse minutes.')

    if minutes not in range(0, 60):
      raise ValueError(f'Minutes value: {minutes:d} out of bounds.')

    try:
      seconds = int(time_string[6:8], 10)
    except ValueError:
      raise ValueError('Unable to parse day of seconds.')

    # TODO: support a leap second?
    if seconds not in range(0, 60):
      raise ValueError(f'Seconds value: {seconds:d} out of bounds.')

    nanoseconds = None
    time_zone_offset = None

    time_zone_string_index = 8
    while time_zone_string_index < time_string_length:
      if time_string[time_zone_string_index] in ('+', '-'):
        break

      time_zone_string_index += 1

    # The calculations that follow rely on the time zone string index
    # to point beyond the string in case no time zone offset was defined.
    if time_zone_string_index == time_string_length - 1:
      time_zone_string_index += 1

    if time_string_length > 8 and time_string[8] == '.':
      time_fraction_length = time_zone_string_index - 9
      if time_fraction_length not in (3, 6, 9):
        raise ValueError('Invalid time string.')

      try:
        time_fraction = time_string[9:time_zone_string_index]
        time_fraction = int(time_fraction, 10)
      except ValueError:
        raise ValueError('Unable to parse time fraction.')

      if time_fraction_length == 3:
        time_fraction *= 1000000
      elif time_fraction_length == 6:
        time_fraction *= 1000

      nanoseconds = time_fraction

    if time_zone_string_index < time_string_length:
      if (time_string_length - time_zone_string_index != 6 or
          time_string[time_zone_string_index + 3] != ':'):
        raise ValueError('Invalid time string.')

      try:
        hours_from_utc = int(time_string[
            time_zone_string_index + 1:time_zone_string_index + 3])
      except ValueError:
        raise ValueError('Unable to parse time zone hours offset.')

      if hours_from_utc not in range(0, 15):
        raise ValueError('Time zone hours offset value out of bounds.')

      try:
        minutes_from_utc = int(time_string[
            time_zone_string_index + 4:time_zone_string_index + 6])
      except ValueError:
        raise ValueError('Unable to parse time zone minutes offset.')

      if minutes_from_utc not in range(0, 60):
        raise ValueError('Time zone minutes offset value out of bounds.')

      # pylint: disable=invalid-unary-operand-type
      time_zone_offset = (hours_from_utc * 60) + minutes_from_utc

      if time_string[time_zone_string_index] == '-':
        time_zone_offset = -time_zone_offset

    return hours, minutes, seconds, nanoseconds, time_zone_offset

  def _GetDateValues(
      self, number_of_days, epoch_year, epoch_month, epoch_day_of_month):
    """Determines date values.

    Args:
      number_of_days (int): number of days since epoch.
      epoch_year (int): year that is the start of the epoch e.g. 1970.
      epoch_month (int): month that is the start of the epoch, where
          1 represents January.
      epoch_day_of_month (int): day of month that is the start of the epoch,
          where 1 represents the first day.

    Returns:
       tuple[int, int, int]: year, month, day of month.

    Raises:
      ValueError: if the epoch year, month or day of month values are out
          of bounds.
    """
    if epoch_year < 0:
      raise ValueError(f'Epoch year value: {epoch_year:d} out of bounds.')

    if epoch_month not in range(1, 13):
      raise ValueError(f'Epoch month value: {epoch_month:d} out of bounds.')

    epoch_days_per_month = self._GetDaysPerMonth(epoch_year, epoch_month)
    if epoch_day_of_month < 1 or epoch_day_of_month > epoch_days_per_month:
      raise ValueError(
          f'Epoch day of month value: {epoch_day_of_month:d} out of bounds.')

    before_epoch = number_of_days < 0

    year = epoch_year
    month = epoch_month
    if before_epoch:
      month -= 1
      if month <= 0:
        month = 12
        year -= 1

    number_of_days += epoch_day_of_month
    if before_epoch:
      number_of_days *= -1

    # Align with the start of the year.
    while month > 1:
      days_per_month = self._GetDaysPerMonth(year, month)
      if number_of_days < days_per_month:
        break

      if before_epoch:
        month -= 1
      else:
        month += 1

      if month > 12:
        month = 1
        year += 1

      number_of_days -= days_per_month

    # Align with the start of the next century.
    _, remainder = divmod(year, 100)
    for _ in range(remainder, 100):
      days_in_year = self._GetNumberOfDaysInYear(year)
      if number_of_days < days_in_year:
        break

      if before_epoch:
        year -= 1
      else:
        year += 1

      number_of_days -= days_in_year

    days_in_century = self._GetNumberOfDaysInCentury(year)
    while number_of_days > days_in_century:
      if before_epoch:
        year -= 100
      else:
        year += 100

      number_of_days -= days_in_century
      days_in_century = self._GetNumberOfDaysInCentury(year)

    days_in_year = self._GetNumberOfDaysInYear(year)
    while number_of_days > days_in_year:
      if before_epoch:
        year -= 1
      else:
        year += 1

      number_of_days -= days_in_year
      days_in_year = self._GetNumberOfDaysInYear(year)

    days_per_month = self._GetDaysPerMonth(year, month)
    while number_of_days > days_per_month:
      if before_epoch:
        month -= 1
      else:
        month += 1

      if month <= 0:
        month = 12
        year -= 1
      elif month > 12:
        month = 1
        year += 1

      number_of_days -= days_per_month
      days_per_month = self._GetDaysPerMonth(year, month)

    if before_epoch:
      days_per_month = self._GetDaysPerMonth(year, month)
      number_of_days = days_per_month - number_of_days

    elif number_of_days == 0:
      number_of_days = 31
      month = 12
      year -= 1

    return year, month, number_of_days

  def _GetDateValuesWithEpoch(self, number_of_days, date_time_epoch):
    """Determines date values.

    Args:
      number_of_days (int): number of days since epoch.
      date_time_epoch (DateTimeEpoch): date and time of the epoch.

    Returns:
       tuple[int, int, int]: year, month, day of month.
    """
    return self._GetDateValues(
        number_of_days, date_time_epoch.year, date_time_epoch.month,
        date_time_epoch.day_of_month)

  def _GetDateWithTimeOfDay(self):
    """Retrieves the date with time of day.

    Note that the date and time are adjusted to UTC.

    Returns:
       tuple[int, int, int, int, int, int]: year, month, day of month, hours,
           minutes, seconds or (None, None, None, None, None, None)
           if the date and time values do not represent a date or time of day.
    """
    normalized_timestamp = self._GetNormalizedTimestamp()
    if normalized_timestamp is None:
      return None, None, None, None, None, None

    if (not self._cached_date_time_values or
        self._cached_date_time_values[0] != normalized_timestamp):
      number_of_days, hours, minutes, seconds = self._GetTimeValues(
          normalized_timestamp)

      try:
        year, month, day_of_month = self._GetDateValuesWithEpoch(
            number_of_days, self._EPOCH_NORMALIZED_TIME)

      except ValueError:
        return None, None, None, None, None, None

      self._cached_date_time_values = (
          normalized_timestamp, year, month, day_of_month, hours, minutes,
          seconds)

    return self._cached_date_time_values[1:]

  def _GetDayOfYear(self, year, month, day_of_month):
    """Retrieves the day of the year for a specific day of a month in a year.

    Args:
      year (int): year e.g. 1970.
      month (int): month, where 1 represents January.
      day_of_month (int): day of the month, where 1 represents the first day.

    Returns:
      int: day of year.

    Raises:
      ValueError: if the month or day of month value is out of bounds.
    """
    if month not in range(1, 13):
      raise ValueError('Month value out of bounds.')

    days_per_month = self._GetDaysPerMonth(year, month)
    if day_of_month < 1 or day_of_month > days_per_month:
      raise ValueError('Day of month value out of bounds.')

    day_of_year = day_of_month
    for past_month in range(1, month):
      day_of_year += self._GetDaysPerMonth(year, past_month)

    return day_of_year

  def _GetDaysPerMonth(self, year, month):
    """Retrieves the number of days in a month of a specific year.

    Args:
      year (int): year e.g. 1970.
      month (int): month, where 1 represents January.

    Returns:
      int: number of days in the month.

    Raises:
      ValueError: if the month value is out of bounds.
    """
    if month not in range(1, 13):
      raise ValueError('Month value out of bounds.')

    days_per_month = definitions.DAYS_PER_MONTH[month - 1]
    if month == 2 and (self._is_delta or self._IsLeapYear(year)):
      days_per_month += 1

    return days_per_month

  @abc.abstractmethod
  def _GetNormalizedTimestamp(self):
    """Retrieves the normalized timestamp.

    Returns:
      decimal.Decimal: normalized timestamp, which contains the number of
          seconds since January 1, 1970 00:00:00 and a fraction of second used
          for increased precision, or None if the normalized timestamp cannot be
          determined.
    """

  def _GetNumberOfDaysInCentury(self, year):
    """Retrieves the number of days in a century.

    Args:
      year (int): year in the century e.g. 1970.

    Returns:
      int: number of (remaining) days in the century.

    Raises:
      ValueError: if the year value is out of bounds.
    """
    if year < 0:
      raise ValueError('Year value out of bounds.')

    year, _ = divmod(year, 100)
    year *= 100

    number_of_days = definitions.DAYS_PER_CENTURY.get(year, None)
    if number_of_days is not None:
      return number_of_days

    if self._IsLeapYear(year):
      return 36525
    return 36524

  def _GetNumberOfDaysInYear(self, year):
    """Retrieves the number of days in a specific year.

    Args:
      year (int): year e.g. 1970.

    Returns:
      int: number of days in the year.
    """
    number_of_days = definitions.DAYS_PER_YEAR.get(year, None)
    if number_of_days is not None:
      return number_of_days

    if self._IsLeapYear(year):
      return 366
    return 365

  def _GetNumberOfSecondsFromElements(
      self, year, month, day_of_month, hours, minutes, seconds):
    """Retrieves the number of seconds from the date and time elements.

    Args:
      year (int): year e.g. 1970.
      month (int): month, where 1 represents January.
      day_of_month (int): day of the month, where 1 represents the first day.
      hours (int): hours.
      minutes (int): minutes.
      seconds (int): seconds.

    Returns:
      int: number of seconds since January 1, 1970 00:00:00 or None if year,
          month or day of month are not set.

    Raises:
      ValueError: if the time elements are invalid.
    """
    if not month or not day_of_month:
      return None

    if hours is None:
      hours = 0
    elif hours not in range(0, 24):
      raise ValueError(f'Hours value: {hours!s} out of bounds.')

    if minutes is None:
      minutes = 0
    elif minutes not in range(0, 60):
      raise ValueError(f'Minutes value: {minutes!s} out of bounds.')

    # TODO: support a leap second?
    if seconds is None:
      seconds = 0
    elif seconds not in range(0, 60):
      raise ValueError(f'Seconds value: {seconds!s} out of bounds.')

    number_of_days = definitions.DAYS_PER_YEAR_IN_POSIX_EPOCH.get(year, None)
    if number_of_days is None:
      raise ValueError(f'Year value: {year!s} out of bounds.')

    number_of_days += sum(
        definitions.DAYS_PER_MONTH[index] for index in range(month - 1))
    if month > 2 and self._IsLeapYear(year):
      number_of_days += 1

    days_per_month = self._GetDaysPerMonth(year, month)
    if day_of_month < 1 or day_of_month > days_per_month:
      raise ValueError(f'Day of month value: {day_of_month:d} out of bounds.')

    number_of_days += day_of_month - 1
    number_of_hours = (number_of_days * 24) + hours
    number_of_minutes = (number_of_hours * 60) + minutes
    number_of_seconds = (number_of_minutes * 60) + seconds

    return number_of_seconds

  def _GetTimeValues(self, number_of_seconds):
    """Determines time values.

    Args:
      number_of_seconds (int|decimal.Decimal): number of seconds.

    Returns:
       tuple[int, int, int, int]: days, hours, minutes, seconds.
    """
    number_of_seconds = int(number_of_seconds)
    number_of_minutes, seconds = divmod(number_of_seconds, 60)
    number_of_hours, minutes = divmod(number_of_minutes, 60)
    number_of_days, hours = divmod(number_of_hours, 24)
    return number_of_days, hours, minutes, seconds

  def _IsLeapYear(self, year):
    """Determines if a year is a leap year.

    Args:
      year (int): year e.g. 1970.

    Returns:
      bool: True if the year is a leap year.
    """
    # pylint: disable=consider-using-ternary
    return (year % 4 == 0 and year % 100 != 0) or year % 400 == 0

  @abc.abstractmethod
  def CopyFromDateTimeString(self, time_string):
    """Copies a date time value from a date and time string.

    Args:
      time_string (str): date and time value formatted as:
          YYYY-MM-DD hh:mm:ss.######[+-]##:##

          Where # are numeric digits ranging from 0 to 9 and the seconds
          fraction can be either 3, 6 or 9 digits. The time of day, seconds
          fraction and time zone offset are optional. The default time zone
          is UTC.

    Raises:
      ValueError: if the time string is invalid or not supported.
    """

  def CopyToPosixTimestamp(self):
    """Copies the date time value to a POSIX timestamp.

    Returns:
      int: a POSIX timestamp in seconds or None if no timestamp is available.
    """
    normalized_timestamp = self._GetNormalizedTimestamp()
    if normalized_timestamp is None:
      return None

    return int(normalized_timestamp)

  def CopyToPosixTimestampWithFractionOfSecond(self):
    """Copies the date time value to a POSIX timestamp with fraction of second.

    Returns:
      tuple[int, int]: a POSIX timestamp in seconds with fraction of second or
          None, None if no timestamp is available.
    """
    normalized_timestamp = self._GetNormalizedTimestamp()
    if normalized_timestamp is None:
      return None, None

    remainder_multiplier = self._REMAINDER_MULTIPLIER.get(self._precision, None)
    if not remainder_multiplier:
      remainder = None
    elif normalized_timestamp >= 0:
      remainder = int((normalized_timestamp % 1) * remainder_multiplier)
    else:
      remainder = int((normalized_timestamp % 1) * -remainder_multiplier)

    return int(normalized_timestamp), remainder

  @abc.abstractmethod
  def CopyToDateTimeString(self):
    """Copies the date time value to a date and time string.

    Returns:
      str: date and time value formatted as: "YYYY-MM-DD hh:mm:ss.######" or
          None if the timestamp cannot be copied to a date and time string.
    """

  def CopyToDateTimeStringISO8601(self):
    """Copies the date time value to an ISO 8601 date and time string.

    Returns:
      str: date and time value formatted as an ISO 8601 date and time string or
          None if the timestamp cannot be copied to a date and time string.
    """
    date_time_string = self.CopyToDateTimeString()
    if date_time_string:
      date_time_string = date_time_string.replace(' ', 'T')

      if self._time_zone_offset is not None or not self.is_local_time:
        time_zone_offset_hours, time_zone_offset_minutes = divmod(
            self._time_zone_offset or 0, 60)
        if time_zone_offset_hours >= 0:
          time_zone_offset_sign = '+'
        else:
          time_zone_offset_sign = '-'
          time_zone_offset_hours *= -1

        time_zone_string = (
            f'{time_zone_offset_hours:02d}:{time_zone_offset_minutes:02d}')
        date_time_string = time_zone_offset_sign.join([
            date_time_string, time_zone_string])

    return date_time_string

  def GetDate(self):
    """Retrieves the date represented by the date and time values.

    Note that the date is adjusted to UTC.

    Returns:
       tuple[int, int, int]: year, month, day of month or (None, None, None)
           if the date and time values do not represent a date.
    """
    year, month, day_of_month, _, _, _ = self._GetDateWithTimeOfDay()
    return year, month, day_of_month

  def GetDateWithTimeOfDay(self):
    """Retrieves the date with time of day.

    Note that the date and time are adjusted to UTC.

    Returns:
       tuple[int, int, int, int, int, int]: year, month, day of month, hours,
           minutes, seconds or (None, None, None, None, None, None)
           if the date and time values do not represent a date or time of day.
    """
    return self._GetDateWithTimeOfDay()

  # TODO: remove this method when there is no more need for it in Plaso.
  def GetPlasoTimestamp(self):
    """Retrieves a timestamp that is compatible with Plaso.

    Returns:
      int: a POSIX timestamp in microseconds or None if no timestamp is
          available.

    Raises:
      ValueError: if the timestamp cannot be determined.
    """
    normalized_timestamp = self._GetNormalizedTimestamp()
    if normalized_timestamp is None:
      return None

    normalized_timestamp *= definitions.MICROSECONDS_PER_SECOND

    try:
      normalized_timestamp = normalized_timestamp.quantize(
          1, rounding=decimal.ROUND_HALF_UP)
    except decimal.InvalidOperation as exception:
      raise ValueError(
          f'Unable to round normalized timestamp with error: {exception!s}')

    return int(normalized_timestamp)

  def GetTimeOfDay(self):
    """Retrieves the time of day represented by the date and time values.

    Note that the time is adjusted to UTC.

    Returns:
       tuple[int, int, int]: hours, minutes, seconds or (None, None, None)
           if the date and time values do not represent a time of day.
    """
    _, _, _, hours, minutes, seconds = self._GetDateWithTimeOfDay()
    return hours, minutes, seconds



# -*- coding: utf-8 -*-
"""The date and time definitions.

Also see:
  https://en.wikipedia.org/wiki/Day
  https://en.wikipedia.org/wiki/Hour
  https://en.wikipedia.org/wiki/Minute
"""

DAYS_PER_MONTH = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

SECONDS_PER_DAY = 86400

DECISECONDS_PER_SECOND = 10

CENTISECONDS_PER_SECOND = 100

MILLISECONDS_PER_SECOND = 1000

DECIMICROSECONDS_PER_SECOND = 10000

MICROSECONDS_PER_DAY = 86400000000
MICROSECONDS_PER_SECOND = 1000000
MICROSECONDS_PER_DECISECOND = 100000
MICROSECONDS_PER_MILLISECOND = 1000

NANOSECONDS_PER_DAY = 86400000000000
NANOSECONDS_PER_SECOND = 1000000000
NANOSECONDS_PER_DECISECOND = 100000000
NANOSECONDS_PER_CENTISECOND = 10000000
NANOSECONDS_PER_MILLISECOND = 1000000
NANOSECONDS_PER_DECIMILISECOND = 100000
NANOSECONDS_PER_MICROSECOND = 1000

PRECISION_1_DAY = '1d'
PRECISION_1_HOUR = '1h'
PRECISION_1_NANOSECOND = '1ns'
PRECISION_10_NANOSECONDS = '10s'
PRECISION_100_NANOSECONDS = '100ns'
PRECISION_1_MICROSECOND = '1us'
PRECISION_10_MICROSECONDS = '10us'
PRECISION_100_MICROSECONDS = '100us'
PRECISION_1_MILLISECOND = '1ms'
PRECISION_10_MILLISECONDS = '10ms'
PRECISION_100_MILLISECONDS = '100ms'
PRECISION_1_MINUTE = '1min'
PRECISION_1_SECOND = '1s'
PRECISION_2_SECONDS = '2s'

PRECISION_VALUES = frozenset([
    PRECISION_1_DAY,
    PRECISION_1_HOUR,
    PRECISION_1_NANOSECOND,
    PRECISION_10_NANOSECONDS,
    PRECISION_100_NANOSECONDS,
    PRECISION_1_MICROSECOND,
    PRECISION_10_MICROSECONDS,
    PRECISION_100_MICROSECONDS,
    PRECISION_1_MILLISECOND,
    PRECISION_10_MILLISECONDS,
    PRECISION_100_MILLISECONDS,
    PRECISION_1_MINUTE,
    PRECISION_1_SECOND,
    PRECISION_2_SECONDS])

# Create a days per century lookup table.
DAYS_PER_CENTURY = {}
for year in range(-10000, 10000, 100):
  if (year % 4 == 0 and year % 100 != 0) or year % 400 == 0:
    number_of_days = 36525
  else:
    number_of_days = 36524
  DAYS_PER_CENTURY[year] = number_of_days

# Create a days per year lookup table.
DAYS_PER_YEAR = {}
for year in range(-10000, 10000, 1):
  if (year % 4 == 0 and year % 100 != 0) or year % 400 == 0:
    number_of_days = 366
  else:
    number_of_days = 365
  DAYS_PER_YEAR[year] = number_of_days

# Create a days per year in POSIX epoch lookup table.
DAYS_PER_YEAR_IN_POSIX_EPOCH = {}

number_of_days = 0
for year in range(1969, -10000, -1):
  number_of_days -= DAYS_PER_YEAR[year]
  DAYS_PER_YEAR_IN_POSIX_EPOCH[year] = number_of_days

number_of_days = 0
for year in range(1970, 10000, 1):
  DAYS_PER_YEAR_IN_POSIX_EPOCH[year] = number_of_days
  number_of_days += DAYS_PER_YEAR[year]
